"""
mmg_toolbox tests
Test tkinter dataviewer
"""

import tkinter as tk

from mmg_toolbox.tkguis.misc.styles import create_root
from mmg_toolbox.tkguis.misc.config import default_config, C
from mmg_toolbox.tkguis.widgets.nexus_data_viewer import NexusDataViewer
from . import only_dls_file_system


def test_config():
    config = default_config(beamline='')
    assert C.conf_file in config
    assert C.metadata_string in config
    assert C.default_metadata in config

    config = default_config(beamline='i10-1')
    assert config.get(C.beamline) == 'i10-1'


@only_dls_file_system
def test_widgets():
    config = default_config()

    from mmg_toolbox.tkguis.widgets.title_window import TitleWindow
    root = create_root('test')
    widget = TitleWindow(root, config)
    assert widget
    root.destroy()

    from mmg_toolbox.tkguis.widgets.multi_scan_analysis import MultiScanAnalysis
    root = create_root('test')
    widget = MultiScanAnalysis(root, config)
    assert widget
    root.destroy()

    from mmg_toolbox.tkguis.widgets.peak_fit_analysis import PeakFitAnalysis
    root = create_root('test')
    widget = PeakFitAnalysis(root, config)
    assert widget
    root.destroy()


@only_dls_file_system
def test_data_viewer():
    f = '/dls/science/groups/das/ExampleData/i16/azimuths'

    root = create_root('test')
    config = default_config(beamline='i16')
    widget = NexusDataViewer(root, f, config)
    assert widget
    # Check initial line plotted
    widget.selector_widget.add_folder(f)
    widget.select_first_file()
    widget.on_file_select()
    assert len(widget.plot_widget.plot_list) == 1
    assert len(widget.plot_widget.listbox.get_children()) == 21
    s = "cmd = flyscancn eta_fly 0.002 101 merlin 0.5 0.25 mroi2"
    assert s in widget.detail_widget.textbox.get(1.0, tk.END)

    # Check multi-plot
    folder_iid = next(iter(widget.selector_widget.tree.get_children()))
    scan_items = widget.selector_widget.tree.get_children(folder_iid)
    assert len(scan_items) > 100
    widget.selector_widget.tree.selection_set(*scan_items[:5])
    widget.on_file_select()  # <<TreeviewSelect>>
    assert len(widget.plot_widget.plot_list) == 5

    # Check multi-y-axis
    y_axis_items = widget.plot_widget.listbox.get_children()
    assert len(y_axis_items) == 21
    widget.plot_widget.listbox.selection_set(y_axis_items[:2])
    widget.plot_widget.select_listbox_items()  # <<TreeviewSelect>>
    assert len(widget.plot_widget.plot_list) == 10

    # Check fitting
    widget.selector_widget.tree.selection_set(scan_items[0])
    widget.on_file_select()  # <<TreeviewSelect>>
    assert len(widget.plot_widget.plot_list) == 1
    # perform fit
    widget.plot_widget.perform_fit()
    assert widget.plot_widget._fit_result.amplitude > 1
    assert len(widget.plot_widget.listbox.get_children()) == len(y_axis_items) + 1
    # plot fit
    widget.plot_widget.select_listbox_items()
    assert len(widget.plot_widget.plot_list) == 2


@only_dls_file_system
def test_nexus_viewer_rois():
    f = '/dls/science/groups/das/ExampleData/i16'
    root = create_root('test')
    config = default_config()
    widget = NexusDataViewer(root, f, config)
    assert widget
    # Check initial line plotted
    widget.selector_widget.add_folder(f)
    widget.select_first_file()

    # Select scan
    widget.selector_widget.select_box.set('1041304')
    widget.selector_widget.select_from_box()
    widget.on_file_select()  # <<TreeviewSelect>>
    assert abs(widget.plot_widget.index_line.get_xdata()[0] - 72.69) < 0.01

    # Check detector image
    widget.plot_widget.view_index.set(56)
    widget.plot_widget.update_image()
    image = widget.plot_widget.ax_image.get_array()
    assert image.max() > 100000

    # Add new ROI
    widget.plot_widget.add_roi('test_roi', 152, 208, 63, 54, 'pil3_100k')
    y_axis_items = widget.plot_widget.listbox.get_children()
    y_axis_item = next(iid for iid in y_axis_items if 'test_roi_total' in widget.plot_widget.listbox.item(iid, 'text'))
    widget.plot_widget.listbox.selection_set(y_axis_item)
    widget.plot_widget.select_listbox_items()  # <<TreeviewSelect>>
    assert widget.plot_widget.line.get_ydata().max() > 3e8







from .apps.experiment import create_title_window
from .apps.data_viewer import create_data_viewer
from .apps.nexus import create_nexus_viewer
from .apps.file_browser import create_nexus_file_browser
from .cli import cli_run, run

__all__ = [
    'create_nexus_file_browser',
    'create_nexus_viewer',
    'create_data_viewer',
    'create_title_window',
    'cli_run',
    'run'
]




# """
# Try out Spectra
# """
#
# import os
# import matplotlib.pyplot as plt
# from mmg_toolbox.xas.spectra import SpectraScan, find_pol_pairs
# from mmg_toolbox.nexus.nexus_writer import create_xmcd_nexus
#
# if os.path.isdir('/dls'):
#     files = [
#         '/dls/i10-1/data/2025/cm40624-2/i10-1-26822.nxs',
#         '/dls/i10-1/data/2025/cm40624-2/i10-1-26823.nxs',
#         '/dls/i10-1/data/2025/cm40624-2/i10-1-26824.nxs',
#         '/dls/i10-1/data/2025/cm40624-2/i10-1-26825.nxs',
#     ]
# else:
#     files = [
#         r"C:\Users\grp66007\OneDrive - Diamond Light Source Ltd\DataAnalysis\I06\example_xmcd\i10-1-207.nxs",
#         r"C:\Users\grp66007\OneDrive - Diamond Light Source Ltd\DataAnalysis\I06\example_xmcd\i10-1-208.nxs",
#         r"C:\Users\grp66007\OneDrive - Diamond Light Source Ltd\DataAnalysis\I06\example_xmcd\i10-1-209.nxs",
#         r"C:\Users\grp66007\OneDrive - Diamond Light Source Ltd\DataAnalysis\I06\example_xmcd\i10-1-210.nxs",
#         r"C:\Users\grp66007\OneDrive - Diamond Light Source Ltd\DataAnalysis\I06\example_xmcd\i10-1-211.nxs",
#         r"C:\Users\grp66007\OneDrive - Diamond Light Source Ltd\DataAnalysis\I06\example_xmcd\i10-1-212.nxs",
#     ]
#     # files = [
#     #     r"C:\Users\grp66007\OneDrive - Diamond Light Source Ltd\PythonProjects\xmcd_analysis\data\i10-1-26822.nxs",
#     #     r"C:\Users\grp66007\OneDrive - Diamond Light Source Ltd\PythonProjects\xmcd_analysis\data\i10-1-26823.nxs",
#     #     r"C:\Users\grp66007\OneDrive - Diamond Light Source Ltd\PythonProjects\xmcd_analysis\data\i10-1-26824.nxs",
#     #     r"C:\Users\grp66007\OneDrive - Diamond Light Source Ltd\PythonProjects\xmcd_analysis\data\i10-1-26825.nxs",
#     # ]
#
# scans = [SpectraScan(file) for file in files]
#
# print(scans)
#
# from mmg_toolbox.xas.spectra_analysis import fit_double_edge_step_background
#
# scan = scans[0]
# # scan.create_figure()
# signal, bkg = fit_double_edge_step_background(scan.tey.energy, scan.tey.signal, 709, 722, 10)
#
# # pol1 = scans[0] + scans[3]
# # pol2 = scans[1] + scans[2]
# #
# # print(pol1)
# # pol1.create_figure()
# #
# # diff = pol1 - pol2
# #
# # print(diff)
# #
# # diff.create_figure()
#
# """
# pairs = find_pol_pairs(*scans)
#
# for pair in pairs:
#     pair.create_figure()
#
# rem_bkg = [s.remove_background('flat').remove_background('exp').norm_to_jump() for s in scans]
# pairs = find_pol_pairs(*rem_bkg)
#
# for pair in pairs:
#     pair.create_figure()
#
# pol1 = [pair.spectra1 for pair in pairs]
# pol2 = [pair.spectra2 for pair in pairs]
# av_pol1 = sum(pol1[1:], pol1[0])
# av_pol2 = sum(pol2[1:], pol2[0])
# diff = av_pol1 - av_pol2
# diff.create_figure()
# print(diff)
# """
#
# plt.show()


"""
{{title}}
{{description}}
"""

import matplotlib.pyplot as plt
from mmg_toolbox import Experiment

data_dir = '{{experiment_dir}}'
scan_numbers = {{scan_numbers}}

exp = Experiment(data_dir, instrument='{{beamline}}')
exp.plot.set_plot_defaults()

fig, ax = plt.subplots()
exp.plot.multi_lines(
    *scan_numbers,
    xaxis='{{x-axis}}',
    yaxis='{{y-axis}}',
    value='{{value}}',
    axes=ax
)

plt.show()



"""
Test nbrunner
"""

# from mmg_toolbox.utils.nb_runner import process_template, view_jupyter_notebook
#
#
# template = '/dls_sw/i06/scripts/gda-zocalo/notebooks/xas_notebook.ipynb'
# nx_file = '/dls/i10-1/data/2025/cm40624-2/i10-1-26825.nxs'
# out = None #'/scratch/grp66007/'
#
# ipynb, html = process_template(template, nx_file, out)
#
# view_jupyter_notebook(html)




"""
Generic metadata from NeXus files for use by HdfMap
"""

class HdfMapNexus:
    """HdfMap Eval commands for any nexus file"""
    instrument = 'NXinstrument_name?("beamline")'
    sample = 'NXsample_name?("none")'
    date = 'NXentry_start_time'


class HdfMapMMGMetadata(HdfMapNexus):
    """HdfMap Eval commaands for all MMG beamlines"""
    cmd = '(cmd|user_command|scan_command)'
    temp = '(T_sample|Tsample|itc3_device_sensor_temp|lakeshore336_cryostat|lakeshore336_sample?(300))'
    beamline = '(beamline|instrument_name)'
    endstation = '(end_station|instrument_name)'
    field = 'sqrt(field_x?(0)**2 + field_y?(0)**2 + field_z?(0)**2)'
    field_x = 'field_x?(0)'
    field_y = 'field_y?(0)'
    field_z = '(magnet_field|ips_demand_field|field_z?(0))'
    energy = '(fastEnergy|pgm_energy|energye|energyh|energy)'
    pol = 'polarisation?("lh")'


class HdfMapXASMetadata(HdfMapMMGMetadata):
    """HdfMap Eval commands for I06 & I10 metadata"""
    # iddgap = 'idd_gap'
    # rowphase = 'idu_trp if idd_gap == 100 else idd_trp'
    mode = '"tey"'  # currently grabs the last NXdata.mode, not the first
    rot = '(scmth|xabs_theta|ddiff_theta|em_pitch|hfm_pitch?(0))'
    monitor = '(i0|C2|ca62sr|mcs16|macr16|mcse16|macj316|mcsh16|macj216)'
    tey = '(tey|C1|ca61sr|mcs17|macr17|mcse17|macj317|mcsh17|macj217)'
    tfy = '(fdu|C3|ca63sr|mcs18|macr18|mcse18|macj318|mcsh18|macaj218)'




"""
mmg_toolbox tests
Test plotting functions
"""

import numpy as np
from matplotlib import pyplot as plt
from mpl_toolkits.mplot3d.axes3d import Axes3D

import mmg_toolbox.plotting.matplotlib as plots
from mmg_toolbox.utils.experiment import Experiment
from . import only_dls_file_system
from .example_files import DIR


def test_matplotlib():
    n_plots = 25
    fig_ax = plots.generate_subplots(n_plots, subplots=(4, 4), suptitle='test')
    assert len(fig_ax) == n_plots

    x = np.arange(-10, 10, 0.1)
    y = x ** 2
    errors = np.sqrt(y + 10)
    lines = plots.plot_line(fig_ax[0][1], x, y, errors, '+-', label='stuff')
    assert len(lines) == 3

    plot_data = [(n, x, y + n) for n in range(10)]
    lines, sm = plots.plot_lines(fig_ax[1][1],*plot_data)
    assert len(lines) == 10
    assert isinstance(sm, plt.cm.ScalarMappable)


@only_dls_file_system
def test_exp_plots():
    exp = Experiment(DIR + '/i16/cm37262-1', instrument='i16')

    ax = exp.plot(-1)
    assert isinstance(ax, plt.Axes)

    rng = range(1032510, 1032521)
    ax = exp.plot(*rng)
    assert isinstance(ax, plt.Axes)
    fig_ax = exp.plot.multi_plot(*rng)
    fig, ax = fig_ax[0]
    assert isinstance(ax, plt.Axes)
    ax = exp.plot.surface_2d(*rng)
    assert isinstance(ax, plt.Axes)
    ax = exp.plot.lines_3d(*rng)
    assert isinstance(ax, Axes3D)
    ax = exp.plot.surface_3d(*rng)
    assert isinstance(ax, Axes3D)




"""
Menu options for processing tasks
"""
import os
import tkinter as tk

from mmg_toolbox.utils.env_functions import (get_notebook_directory, open_terminal, get_scan_number,
                                             get_processing_directory)
from mmg_toolbox.scripts.scripts import (generate_script, create_notebook,
                                         SCRIPTS, NOTEBOOKS, R)
from .config import get_config, C
from .functions import check_new_file
from .jupyter import launch_jupyter_notebook, terminate_notebooks
from ..apps.multi_scan_analysis import create_multi_scan_analysis
from ..apps.python_editor import create_python_editor


def create_script_from_template(root: tk.Misc, template: str = 'example', directory: str | None = None,
                                config: dict | None = None, **replacements):
    proc_dir = directory or config[C.current_proc]
    script_name = os.path.join(proc_dir, f"{template}.py")
    new_file = check_new_file(root, script_name)
    script = generate_script(template, **replacements)
    create_python_editor(script, root, config, filename=new_file)


def create_notebook_from_template(root: tk.Misc, template: str = 'example', directory: str | None = None,
                                  config: dict | None = None, **replacements):
    proc_dir = directory or config[C.current_proc]
    script_name = os.path.join(proc_dir, f"{template}.ipynb")
    new_file = check_new_file(root, script_name)
    create_notebook(new_file, template, **replacements)
    launch_jupyter_notebook('notebook', file=new_file)


def generate_replacement_getter(scanno_getter, title_getter, x_getter = None,
                                y_getter = None, metadata_getter = None):
    def getter() -> dict:
        replacements = {
            R.scannos: str(scanno_getter()),
            R.description: 'Created by  MultiScanAnalysis',
        }
        if title_getter:
            replacements[R.title] = str(title_getter())
        if x_getter:
            replacements[R.xaxis] = str(x_getter())
        if y_getter:
            replacements[R.yaxis] = str(y_getter())
        if metadata_getter:
            replacements[R.value] = str(metadata_getter())
        return replacements
    return getter


def generate_processing_menu(parent, config: dict, directory: str | None = None, scan_files_getter=None,
                             replacement_getter=None) -> dict:
    """Generate processing menu options"""

    directory = directory or config[C.current_dir]
    proc_dir = config.get(C.current_proc) or get_processing_directory(directory)
    nb_dir = config.get(C.current_nb) or get_notebook_directory(directory)

    if scan_files_getter is None:
        scan_files_getter = lambda: []
    if replacement_getter is None:
        replacement_getter = lambda: {}

    def start_multi_scan_plot():
        filenames = scan_files_getter()
        scan_numbers = [get_scan_number(f) for f in filenames]
        create_multi_scan_analysis(parent, config, exp_directory=directory, scan_numbers=scan_numbers)

    def replacements():
        values = {
            R.exp: directory,
            R.proc: proc_dir,
            R.filepaths: ', '.join(f"'{f}'" for f in scan_files_getter()),
            R.beamline: config.get(C.beamline, ''),
            R.value: config.get(C.default_metadata, ''),
        }
        values.update(replacement_getter())
        return values

    scripts = {
        name: lambda n=name: create_script_from_template(parent, n, directory, config, **replacements())
        for name in SCRIPTS
    }
    notebooks = {
        name: lambda n=name: create_notebook_from_template(parent, n, directory, config, **replacements())
        for name in NOTEBOOKS
    }
    menu = {
        'Multi-Scan': start_multi_scan_plot,
        'Script Editor': lambda: create_python_editor(None, parent, config),
        'Open a terminal': lambda: open_terminal(f"cd {directory}"),
        'Start Jupyter (processing)': lambda: launch_jupyter_notebook('notebook', proc_dir),
        'Start Jupyter (notebooks)': lambda: launch_jupyter_notebook('notebook', nb_dir),
        'Stop Jupyter servers': terminate_notebooks,
        'Create Script:': scripts,
        'Create Notebook:': notebooks,
    }
    return menu



"""
Jupyter notebook runner

From /dls_sw/i16/software/python/jupyter_processor/jproc.py
"""

import os
import shutil
import nbformat
import webbrowser
from nbconvert.preprocessors import ExecutePreprocessor
from nbconvert import HTMLExporter

from .env_functions import run_command, TMPDIR


def read_notebook(filename: str) -> nbformat.NotebookNode:
    return nbformat.read(filename, as_version=4)


def save_notebook(nb: nbformat.NotebookNode, filename: str):
    with open(filename, 'w', encoding='utf-8') as f:
        nbformat.write(nb, f)


def add_code_cell(nb: nbformat.NotebookNode, source: str, index=-1):
    cell = nbformat.v4.new_code_cell(source)
    nb.cells.insert(index, cell)


def add_inpath(nb: nbformat.NotebookNode, filename: str):
    """Add inpath=filename to top of notebook"""
    code = f"inpath = '{filename}'"
    first_cell = next(iter(nb.cells), {})
    first_cell_type = first_cell.get('cell_type', '')
    first_cell_source = first_cell.get('source', '')
    if first_cell_type == 'code' and 'inpath' in first_cell_source:
        # replace contents of cell
        first_cell['source'] = code
    else:
        # add a new cell
        add_code_cell(nb, code, index=0)


def html_processor(nb: nbformat.NotebookNode) -> tuple[str, dict]:
    """
    Process notebook and return (html, resources)
    """
    processor = ExecutePreprocessor()
    html_exporter = HTMLExporter()
    print('\n'.join(str(cell) for cell in nb.cells))
    processor.preprocess(nb, resources={'metadata': {}})
    return html_exporter.from_notebook_node(nb)


def generate_notebook_html(nb: nbformat.NotebookNode) -> tuple[str, dict]:
    """
    Generate html of the notebook without running it
     return (html, resources)
    """
    html_exporter = HTMLExporter()
    return html_exporter.from_notebook_node(nb)


def view_notebook_html(notebook_filename: str):
    """
    Convert a notebook to html and open it in a web-browser
    """
    out_html = os.path.join(TMPDIR, 'tmp_notebook.ipynb')
    nb = read_notebook(notebook_filename)
    (body, resources) = generate_notebook_html(nb)
    with open(out_html, 'w') as f:
        f.write(body)
    webbrowser.open_new_tab(notebook_filename)


def process_template(template: str, nexus_filename: str, output_folder: str | None = None) -> tuple[str, str]:
    """
    Process jupyter notebook
    """
    nb = read_notebook(template)
    add_inpath(nb, nexus_filename)

    if output_folder is None:
        output_folder = TMPDIR
    out_name, ext = os.path.splitext(output_folder)
    if ext:
        out_html = out_name + '.html'
        out_pynb = out_name + '.ipynb'
    else:
        tmp_name, _ = os.path.splitext(os.path.basename(template))
        nxs_name, _ = os.path.splitext(os.path.basename(nexus_filename))
        out_name = f"{nxs_name}_{tmp_name}"
        out_html = os.path.join(output_folder, out_name + '.html')
        out_pynb = os.path.join(output_folder, out_name + '.ipynb')

    # run notebook
    (body, resources) = html_processor(nb)

    with open(out_html, 'w') as f:
        f.write(body)
    save_notebook(nb, out_pynb)
    print(f"Completed notebook, saved as: {out_pynb}")
    return out_pynb, out_html


def run_jupyter_notebook(notebook_filename: str):
    """
    Run a jupyter notebook
    """
    command = f"jupyter notebook {notebook_filename}"
    run_command(command)


def view_jupyter_notebook(notebook_filename: str):
    """
    Open a jupyter notebook.ipynb or .html file
    """
    if notebook_filename.endswith('.ipynb'):
        run_jupyter_notebook(notebook_filename)
    elif notebook_filename.endswith('.html'):
        webbrowser.open_new_tab(notebook_filename)


def reprocess_notebook(notebook_filename: str, output_folder: str | None = None):
    """
    Copy notebook and open jupyter for reprocessing of copied notebook in processing directory
    """
    path, name = os.path.split(notebook_filename)

    if output_folder is None and path.endswith('/processed/notebooks'):
        output_folder = path.replace('/processed/notebooks', '/processing')

    if output_folder.endswith('.ipynb'):
        new_filename = output_folder
    else:
        new_filename = os.path.join(output_folder, name)

    if os.path.isfile(new_filename):
        print(f"{new_filename} already exists, it won't be overwritten")
    else:
        print(f"Creating {new_filename}")
        shutil.copy(notebook_filename, new_filename)
    run_jupyter_notebook(new_filename)




"""
Instrument Model from NeXus files

Makes use of NeXus classes and NXtransformations to build a structure of the instrument from a scan file.
"""

import h5py
import numpy as np

from mmg_toolbox.nexus import nexus_names as nn
from mmg_toolbox.nexus.nexus_functions import nx_find, get_dataset_value
from mmg_toolbox.nexus.nexus_transformations import nx_direction, nx_transformations_max_size, \
    nx_transformations_matrix, nx_transform_vector
from mmg_toolbox.utils.rotations import norm_vector, transform_by_t_matrix
from mmg_toolbox.utils.xray_utils import photon_energy, photon_wavelength
from mmg_toolbox.diffraction.lattice import wavevector, bmatrix
from mmg_toolbox.plotting.matplotlib import Axes3D

# types
Shape = tuple[int, int, int]  #  (n, i, j) == (frame, slow_pixel, fast_pixel) #TODO: should this be fast, slow?
Pixel = tuple[int, float, float]  # (n, i, j) pixel coordinates

#TODO: add polarisation
#TODO: improve docs

class NXBeam:
    """
    NXbeam object
    """

    def __init__(self, path: str, hdf_file: h5py.File):
        self.file = hdf_file
        self.path = path
        self.beam = hdf_file[path]

        self.direction = nx_direction(path, hdf_file)
        self.en, self.wl = self.energy_wavelength()
        self.wv = wavevector(self.wl)
        self.incident_wavevector = self.wv * self.direction

    def energy_wavelength(self):
        """
        Return beam energy in keV and wavelength in A
        :return: incident_energy, incident_wavelength
        """
        if nn.NX_WL in self.beam:
            wl = get_dataset_value(nn.NX_WL, self.beam, units='A')
            return photon_energy(wl), wl
        elif nn.NX_EN in self.beam:
            en = get_dataset_value(nn.NX_EN, self.beam, units='keV')
            return en, photon_wavelength(en)
        else:
            raise KeyError(f"{self.beam} contains no '{nn.NX_WL}' or '{nn.NX_EN}'")

    def __repr__(self):
        return f"NXBeam({self.beam})"


class NXSsample:
    """
    NXsample object
    """

    def __init__(self, path: str, hdf_file: h5py.File):
        self.file = hdf_file
        self.path = path
        self.sample = hdf_file[path]

        self.name = get_dataset_value(nn.NX_NAME, self.sample, 'none')
        self.unit_cell = get_dataset_value(nn.NX_SAMPLE_UC, self.sample, np.array([1., 1, 1, 90, 90, 90]))
        self.orientation_matrix = get_dataset_value(nn.NX_SAMPLE_OM, self.sample, np.eye(3))
        self.ub_matrix = get_dataset_value(nn.NX_SAMPLE_UB, self.sample, bmatrix(*self.unit_cell))

        self.size = nx_transformations_max_size(path, hdf_file)
        self.transforms = [
            nx_transformations_matrix(path, n, hdf_file)
            for n in range(self.size)
        ]  # list of 4x4 transformation matrices

    def __repr__(self):
        return f"NXSsample({self.sample})"

    def hkl2q(self, hkl: tuple[float, float, float] | np.ndarray, index: int = 0) -> np.ndarray:
        """
        Returns wavevector direction for given hkl in lab space
        :param hkl: Miller indices (h, k, l), in units of reciprocal lattice vectors
        :param index: index of scan
        :return: Q position in inverse Angstroms
        """
        hkl = np.reshape(hkl, (-1, 3))
        z = self.transforms[index][:3, :3]
        ub = 2 * np.pi * self.ub_matrix
        return np.dot(z, np.dot(ub, hkl.T)).T


class NXDetectorModule:
    """
    NXdetector_module object
    """

    def __init__(self, path: str, hdf_file: h5py.File):
        self.file = hdf_file
        self.path = path
        self.module = hdf_file[path]

        self.data_origin = get_dataset_value(nn.NX_MODULE_ORIGIN, self.module, np.array([0, 0]))
        self.data_size = get_dataset_value(nn.NX_MODULE_SIZE, self.module, np.array([1, 1]))

        self.module_offset_path = f"{self.path}/{nn.NX_MODULE_OFFSET}"
        self.fast_pixel_direction_path = f"{self.path}/{nn.NX_MODULE_FAST}"
        self.slow_pixel_direction_path = f"{self.path}/{nn.NX_MODULE_SLOW}"

        self.size = nx_transformations_max_size(self.module_offset_path, hdf_file)
        self.offset_transforms = [
            nx_transformations_matrix(self.module_offset_path, n, hdf_file)
            for n in range(self.size)
        ]  # list of 4x4 transformation matrices
        self.fast_transforms = [
            nx_transformations_matrix(self.fast_pixel_direction_path, n, hdf_file)
            for n in range(self.size)
        ]  # list of 4x4 transformation matrices
        self.slow_transforms = [
            nx_transformations_matrix(self.slow_pixel_direction_path, n, hdf_file)
            for n in range(self.size)
        ]  # list of 4x4 transformation matrices

    def __repr__(self):
        return f"NXDetectorModule({self.module})"

    def shape(self) -> Shape:
        """
        Return scan shape of module
            (n, i, j)
        Where:
            n = frames in scan
            i = pixels along slow axis
            j = pixels along fast axis
        """
        return self.size, self.data_size[0], self.data_size[1]

    def pixel_wavevector(self, point: Pixel, wavelength_a) -> np.ndarray:
        """
        Return wavevector of pixel in lab frame
        :param point: (n, i, j) == (frame, slow_axis_pixel, fast_axis_pixel)
        :param wavelength_a: wavelength in Angstrom
        :return: [x, y, z] in inverse Angstrom
        """
        return wavevector(wavelength_a) * self.pixel_direction(point)

    def pixel_direction(self, point: Pixel) -> np.ndarray:
        """
        Return direction of pixel in lab frame
        :param point: (n, i, j) == (frame, slow_axis_pixel, fast_axis_pixel)
        :return: [dx, dy, dz] unit vector
        """
        return norm_vector(self.pixel_position(point))

    def pixel_position(self, point: Pixel) -> np.ndarray:
        """
        Return position of pixel (n, i, j) in lab frame
            n = frame in scan
            i = pixel along slow axis
            j = pixel along fast axis
        """
        index, ii, jj = point

        module_origin = transform_by_t_matrix([0, 0, 0], self.offset_transforms[index])
        fast_pixel = transform_by_t_matrix([0, 0, 0], self.fast_transforms[index])
        slow_pixel = transform_by_t_matrix([0, 0, 0], self.slow_transforms[index])

        fast_direction = fast_pixel - module_origin
        slow_direction = slow_pixel - module_origin
        return np.squeeze(ii * slow_direction + jj * fast_direction + module_origin)

    def corners(self, frame: int) -> np.ndarray:
        """return corners of the detector module at this frame in scan"""
        shape = self.shape()
        corners = np.vstack([
            self.pixel_position((frame, 0, 0)),  # module origin
            self.pixel_position((frame, int(shape[1]), 0)),  # module origin + slow pixels
            self.pixel_position((frame, int(shape[1]), int(shape[2]))),  # o + slow + fast
            self.pixel_position((frame, 0, int(shape[2]))),  # o + fast
            self.pixel_position((frame, 0, 0)),  # module origin
        ])
        return corners


class NXDetector:
    """
    NXdetector object
    """

    def __init__(self, path: str, hdf_file: h5py.File):
        self.file = hdf_file
        self.path = path
        self.detector = hdf_file[path]
        self.size = nx_transformations_max_size(path, hdf_file)
        self.position = nx_transform_vector((0, 0, 0), path, self.size // 2, hdf_file).squeeze()

        self.modules = [
            NXDetectorModule(f"{self.path}/{p}", hdf_file)
            for p, obj in self.detector.items()
            if obj.attrs.get(nn.NX_CLASS) == nn.NX_MODULE.encode()
        ]

    def __repr__(self):
        return f"NXDetector({self.detector}) with {len(self.modules)} modules"


class NXInstrumentModel:
    """
    NXInstrumentModel object
    """

    def __init__(self, hdf_file: h5py.File):
        self.file = hdf_file

        self.entry = nx_find(hdf_file, nn.NX_ENTRY)
        self.instrument = nx_find(self.entry, nn.NX_INST)
        if self.instrument is None:
            raise RuntimeError(f"NXInstrumentModel: instrument not found")

        self.detectors = [
            NXDetector(f"{self.instrument.name}/{p}", hdf_file)
            for p, obj in self.instrument.items()
            if obj.attrs.get(nn.NX_CLASS) == nn.NX_DET.encode()
        ]
        self.components = [
            obj for obj in self.instrument.values()
            if isinstance(obj, h5py.Group) and 'depends_on' in obj
        ]
        self.component_positions = {
            obj.name.split('/')[-1]: nx_transform_vector((0, 0, 0), obj.name, 0, hdf_file).squeeze()
            for obj in self.components
        }
        self.component_positions['sample'] = np.array([0, 0, 0])

        sample_obj = nx_find(self.entry, nn.NX_SAMPLE)
        self.sample = NXSsample(sample_obj.name, hdf_file)
        beam_obj = nx_find(sample_obj, nn.NX_BEAM)
        self.beam = NXBeam(beam_obj.name, hdf_file)

    def __repr__(self):
        return f"NXInstrumentModel({self.file})"

    def _first_detector(self):
        return self.detectors[0].modules[0]

    def shape(self) -> Shape:
        """
        Return scan shape from the first detector module
            (n, i, j)
        Where:
            n = frames in scan
            i = pixels along slow axis
            j = pixels along fast axis
        """
        detector_module = self._first_detector()
        return detector_module.shape()

    def detector_q(self, point: Pixel = (0, 0, 0)) -> np.ndarray:
        """
        return wavevector transfer, kf-ki, of first detector module, in lab frame
        :param point: (n, i, j) == (frame, slow_axis_pixel, fast_axis_pixel)
        :return: [x, y, z] in inverse Angstrom
        """
        wavelength = self.beam.wl
        ki = self.beam.incident_wavevector
        detector_module = self._first_detector()
        kf = detector_module.pixel_wavevector(point, wavelength)
        return kf - ki

    def hkl(self, point: Pixel = (0, 0, 0)):
        """
        return wavevector transfer, Q, of pixel coordinate
        :param point: (n, i, j) == (frame, slow_axis_pixel, fast_axis_pixel)
        :return: [x, y, z] in inverse Angstrom
        """
        n, i, j = point
        q = self.detector_q(point)
        if len(self.sample.transforms) > 1:
            z = self.sample.transforms[n][:3, :3]
        else:
            z = self.sample.transforms[0][:3, :3]  # if sample tranformations aren't stored on every angle
        ub = 2 * np.pi * self.sample.ub_matrix

        inv_ub = np.linalg.inv(ub)
        inv_z = np.linalg.inv(z)

        hphi = np.dot(inv_z, q)
        return np.dot(inv_ub, hphi).T

    def hkl2q(self, hkl: tuple[float, float, float] | np.ndarray):
        """
        Returns wavecector direction for given hkl
        :param hkl: Miller indices, in units of reciprocal lattice vectors
        :return: Q position in inverse Angstroms
        """
        return self.sample.hkl2q(hkl)

    def plot_instrument(self, axes: Axes3D):
        instrument_name = get_dataset_value('name', self.instrument, 'no name')
        max_distance = max([np.linalg.norm(position) for position in self.component_positions.values()])
        max_position = max_distance * self.beam.direction

        axes.plot([-max_position[0], 0], [-max_position[2], 0], [-max_position[1], 0], 'k-')  # beam
        beam_cont = np.linalg.norm(self.detectors[0].position) * self.beam.direction
        axes.plot([0, beam_cont[0]], [0, beam_cont[2]], [0, beam_cont[1]], 'k:')  # continued beam
        # detectors
        for detector in self.detectors:
            pos = detector.position
            axes.plot([0, pos[0]], [0, pos[2]], [0, pos[1]], 'k-')  # scattered beam
        # components
        for component, position in self.component_positions.items():
            axes.plot(position[0], position[2], position[1], 'r+')
            axes.text(position[0], position[2], position[1], s=component)

        axes.set_xlabel('X [mm]')
        axes.set_ylabel('Z [mm]')
        axes.set_zlabel('Y [mm]')
        axes.set_title(f"Instrument: {instrument_name}")
        # ax.set_aspect('equalxz')

    def plot_wavevectors(self, axes: Axes3D):
        frames, ii, jj = self.shape()
        pixel_centre = (frames // 2, ii // 2, jj // 2)
        ki = self.beam.incident_wavevector
        detector_module = self.detectors[0].modules[0]
        kf = detector_module.pixel_wavevector(pixel_centre, self.beam.wl)
        q = kf - ki
        hkl = self.hkl(pixel_centre)

        axes.plot([-ki[0], 0], [-ki[2], 0], [-ki[1], 0], '-k')
        axes.plot([0, kf[0]], [0, kf[2]], [0, kf[1]], '-k')
        axes.plot([0, q[0]], [0, q[2]], [0, q[1]], '-r')

        wl = self.beam.wl
        for frame in range(frames):
            corners = np.vstack([
                detector_module.pixel_wavevector((frame, 0, 0), wl),  # module origin
                detector_module.pixel_wavevector((frame, ii, 0), wl),  # module origin + slow pixels
                detector_module.pixel_wavevector((frame, ii, jj), wl),  # o + slow + fast
                detector_module.pixel_wavevector((frame, 0, jj), wl),  # o + fast
                detector_module.pixel_wavevector((frame, 0, 0), wl),  # module origin
            ])
            axes.plot(corners[:, 0], corners[:, 2], corners[:, 1], '-k')
            corners_q = corners - ki
            axes.plot(corners_q[:, 0], corners_q[:, 2], corners_q[:, 1], '-r')

        # plot Reciprocal lattice
        astar, bstar, cstar = self.hkl2q(np.eye(3))
        axes.plot([0, astar[0]], [0, astar[2]], [0, astar[1]], '-g')
        axes.plot([0, bstar[0]], [0, bstar[2]], [0, bstar[1]], '-g')
        axes.plot([0, cstar[0]], [0, cstar[2]], [0, cstar[1]], '-g')
        axes.text(astar[0], astar[2], astar[1], s='a*')
        axes.text(bstar[0], bstar[2], bstar[1], s='b*')
        axes.text(cstar[0], cstar[2], cstar[1], s='c*')

        axes.set_xlabel(r'Qx [$\AA^{-1}$]', labelpad=20)
        axes.set_ylabel(r'Qz [$\AA^{-1}$]', labelpad=20)
        axes.set_zlabel(r'Qy [$\AA^{-1}$]', labelpad=20)
        axes.set_title(f"Wavevectors\nHKL: {hkl}")
        axes.set_aspect('equal')

    def plot_hkl(self, axes: Axes3D):
        frames, ii, jj = self.shape()
        pixel_centre = (frames // 2, ii // 2, jj // 2)
        hkl = self.hkl(pixel_centre)
        for frame in range(frames):
            corners = np.vstack([
                self.hkl((frame, 0, 0)),  # module origin
                self.hkl((frame, ii, 0)),  # module origin + slow pixels
                self.hkl((frame, ii, jj)),  # o + slow + fast
                self.hkl((frame, 0, jj)),  # o + fast
                self.hkl((frame, 0, 0)),  # module origin
            ])
            axes.plot(corners[:, 0], corners[:, 2], corners[:, 1], '-r')
        origin = self.hkl((0, 0, 0))
        axes.plot(origin[0], origin[2], origin[1], '+k')

        axes.set_xlabel('H')
        axes.set_ylabel('L')
        axes.set_zlabel('K')
        axes.set_title(f"HKL: {hkl}")
        axes.set_aspect('equal')





# Groups
NX_ENTRY = 'NXentry'
NX_DATA = 'NXdata'
NX_INST = 'NXinstrument'
NX_DET = 'NXdetector'
NX_SAMPLE = 'NXsample'
NX_XTL = 'NXcrystal'
NX_MODULE = 'NXdetector_module'
NX_BEAM = 'NXbeam'
NX_SOURCE = 'NXsource'
NX_ID = 'NXinsertion_device'
NX_MONO = 'NXmonochromator'
NX_MON = 'NXmonitor'
NX_NOTE = 'NXnote'
NX_PROC = 'NXprocess'
NX_TRAN = 'NXtransformations'
NX_PARAM = 'NXparameters'

# Fields
NX_WL = 'incident_wavelength'
NX_EN = 'incident_energy'
NX_SHORT_NAME = 'short_name'
NX_NAME = 'name'
NX_SAMPLE_UC = 'unit_cell'
NX_SAMPLE_OM = 'orientation_matrix'
NX_SAMPLE_UB = 'ub_matrix'
NX_STOKES = 'incident_polarization_stokes'
NX_BSIZE = 'extent'
NX_MODULE_ORIGIN = 'data_origin'
NX_MODULE_SIZE = 'data_size'
NX_MODULE_OFFSET = 'module_offset'
NX_MODULE_FAST = 'fast_pixel_direction'
NX_MODULE_SLOW = 'slow_pixel_direction'

# Attributes
NX_CLASS = 'NX_class'
NX_DEFINITION = 'definition'
NX_DEFAULT = 'default'
NX_AXES = 'axes'
NX_SIGNAL = 'signal'
NX_AUXILIARY = 'auxiliary_signals'
NX_INDICES = '{}_indices'
NX_DEPON = 'depends_on'
NX_VECTOR = 'vector'
NX_OFFSET = 'offset'
NX_TTYPE = 'transformation_type'
NX_TROT = 'rotation'
NX_TTRAN = 'translation'
NX_UNITS = 'units'
NX_OFFSET_UNITS = 'offset_units'

# Other
ENTRY_CLASSES = ['NXentry', 'NXsubentry']
XAS_DEFINITIONS = ['NXxas', 'NXxas_new']
SEARCH_ATTRS = (NX_CLASS, 'local_name')  # DLS attribute 'local_name' helps match metadata to scan commands

# Polarisation field names inside NeXus groups
# See https://manual.nexusformat.org/classes/base_classes/NXbeam.html#nxbeam
NX_POLARISATION_FIELDS = [
    'incident_polarization_stokes',  # NXbeam
    'incident_polarization',  # NXbeam
    'polarisation',  # DLS specific in NXinsertion_device
]



"""
Treeview widget for folders
"""

import os
import time
import tkinter as tk
from tkinter import ttk
from threading import Thread, current_thread

import hdfmap

from mmg_toolbox.utils.file_functions import list_files, display_timestamp, get_scan_number
from ..misc.functions import post_right_click_menu, select_folder
from ..misc.logging import create_logger
from ..misc.config import get_config, C
from ..misc.styles import create_hover
from .find_scans import FindScans
from .treeview import CanvasTreeview

logger = create_logger(__file__)


class _ScanSelector(CanvasTreeview):
    """Frame with TreeView for selection of scan files"""

    def __init__(self, root: tk.Misc, config: dict | None = None):
        self.config = config or get_config()
        self.search_str = ""
        self.search_time = time.time()
        self.search_reset = 3.0  # seconds
        self._update_time = 10  # seconds - poll folders for new files

        # Variables
        self.extension = tk.StringVar(root, '.nxs')
        self.read_datasets = tk.BooleanVar(root, True)
        self.search_box = tk.StringVar(root, '')
        self.search_matchcase = tk.BooleanVar(root, False)
        self.search_wholeword = tk.BooleanVar(root, True)
        self.select_box = tk.StringVar(root, '')
        self.map = None

        # Columns
        columns = [
            # (name, text, width, reverse, sort_col)
            ("#0", 'Number', 100, False, None),
            ("modified", 'Date', 150, True, "modified_time"),
            ('modified_time', 'Modified', 0, False, None),
            ("filepath", 'File Path', 0, False, None),
        ]
        # add values from metadata_list
        self.metadata_names = self.config.get(C.metadata_list, {})
        columns += [
            (name, name, 400, True, None) for name in self.metadata_names
        ]
        super().__init__(root, *columns, pack=False)

    "======================================================"
    "=============== populate functions ==================="
    "======================================================"

    def _add_row(self, parent="", name="", timestamp=0.0, time_str="", filepath="", *args, **kwargs):
        values = (time_str, timestamp, filepath) + args
        iid = self.tree.insert(parent, 0, text=name, values=values)
        for name, value in kwargs.items():
            self.tree.set(iid, column=name, value=value)
        return iid

    def _add_file(self, parent, filepath):
        scan_number = str(get_scan_number(filepath))
        timestamp = os.stat(filepath).st_mtime
        mtime = display_timestamp(timestamp)
        iid = self._add_row(parent, name=scan_number, timestamp=timestamp, time_str=mtime, filepath=filepath)
        return iid

    def _add_data(self, item):
        filepath = self.tree.set(item, 'filepath')
        if os.path.isdir(filepath):
            return
        try:
            if self.map is None:
                self.map = hdfmap.create_nexus_map(filepath)
            with hdfmap.load_hdf(filepath) as nxs:
                for name, fmt in self.metadata_names.items():
                    if not self.tree.winfo_exists():
                        return
                    data = self.map.format_hdf(nxs, fmt)
                    self.tree.set(item, name, data)
        except Exception as exception:
            name = next(iter(self.metadata_names), 'data')
            self.tree.set(item, name, str(exception))

    def populate_files(self, item, *file_list: str):
        """Add list of files below folder on folder expand"""
        # remove old entries
        self.tree.delete(*self.tree.get_children(item))
        self.map = None  # reset hdfmap
        start_time = time.time()
        for file in file_list:
            self._add_file(item, file)
        if self.read_datasets.get():
            self.update_metadata()
        logger.info(f"Expanding took {time.time() - start_time:.3g} s")

    def update_metadata(self, event=None):
        """Update dataset values column for hdf files under open folders"""

        def fn():
            for branch in self.tree.get_children():  # folders
                for leaf in self.tree.get_children(branch):  # files
                    if not self.tree.winfo_exists():
                        return
                    self._add_data(leaf)

        th = Thread(target=fn)
        th.daemon = True  # runs thread in the background, outside mainloop, allowing python to close
        th.start()

    def update_files(self):
        """Check folders in the tree for new files"""
        pass

    def poll_files(self, _event=None):
        """Create background thread that checks the folders for new files"""
        def fn():
            while True:
                if not self.tree.winfo_exists():
                    logger.info('poll_files exiting')
                    return
                self.update_files()
                time.sleep(self._update_time)

        th = Thread(target=fn)
        th.daemon = True  # runs thread in the background, outside mainloop, allowing python to close
        th.start()

    "======================================================"
    "================ general functions ==================="
    "======================================================"

    def get_filepath(self) -> tuple[str, str]:
        """
        Return filepath and folderpath of current selection
        :returns filename: str full filepath or None if selection isn't a file
        :returns foldername: str folder path
        """
        filename = None
        foldername = None
        for iid in self.tree.selection():
            filepath = self.tree.set(iid, 'filepath')
            if os.path.isfile(filepath):
                filename = filepath
                foldername = os.path.dirname(filename)
            else:  # item is a folder
                foldername = filepath
        logger.debug(f"Selected item: filename='{filename}', foldername='{foldername}'")
        return filename, foldername

    def get_multi_filepath(self) -> list[str]:
        """
        Return list of filepaths for all files in current selection
        :returns filename: list[str] of full filepaths
        """
        filenames = []
        for iid in self.tree.selection():
            filepath = self.tree.set(iid, 'filepath')
            if os.path.isfile(filepath):
                filenames.append(filepath)
        return filenames

    def copy_path(self):
        filepath, folderpath = self.get_filepath()
        self.root.clipboard_clear()
        if filepath:
            self.root.clipboard_append(filepath)
        else:
            self.root.clipboard_append(folderpath)

    def _create_menu(self, iid: str | int):

        menu = tk.Menu(self.root, tearoff=0)
        menu.add_command(label="Copy path", command=self.copy_path)

        filepath = self.tree.set(iid, 'filepath')
        if os.path.isfile(filepath):
            menu.add_command(label="open Treeview", command=self.open_nexus_treeview)
            menu.add_command(label="open Plot", command=self.open_nexus_plot)
            menu.add_command(label="open Image", command=self.open_nexus_image)
        return menu

    def right_click_menu(self):
        logger.info('Creating right click menu')

        def menu_popup(event):
            # select item
            iid = self.tree.identify_row(event.y)
            if iid:
                menu = self._create_menu(iid)
                post_right_click_menu(menu, event.x_root, event.y_root)
        return menu_popup

    def _delete_tree(self):
        self.tree.delete(*self.tree.get_children())

    def _on_close(self):
        # self.root.unbind_all('<KeyPress>')
        self.root.destroy()

    def _get_select_box(self):
        # TODO: replace with asteval
        return str(eval(self.select_box.get()))

    def select_from_box(self, event=None):
        item = self._get_select_box()
        self.tree.selection_remove(self.tree.selection())
        for iid in self.tree.get_children():  # folders
            for scan_iid in self.tree.get_children(iid):
                scan_number = self.tree.item(scan_iid)['text']
                if item in scan_number:
                    self.tree.selection_add(scan_iid)
                    self.tree.see(scan_iid)
                    break

    def select_box_increase(self):
        if self.tree.selection():
            self.tree.selection_set(self.tree.prev(self.tree.selection()[0]))

    def select_box_decrease(self):
        if self.tree.selection():
            self.tree.selection_set(self.tree.next(self.tree.selection()[0]))

    def on_key_press(self, event):
        """any key press performs search of folders, selects first matching folder"""
        # return if clicked on entry box
        # event.widget == self.tree
        if str(event.widget).endswith('entry'):
            return
        # reset search str after reset time
        ctime = time.time()
        if ctime > self.search_time + self.search_reset:
            self.search_str = ""
        # update search time, add key to query
        self.search_time = ctime
        self.search_str += event.char

        self.tree.selection_remove(self.tree.selection())
        # search folder list
        for branch in self.tree.get_children():  # folders
            folder = self.tree.item(branch)['text']
            if self.search_str in folder[:len(self.search_str)].lower():
            # if folder.lower().startswith(self.search_str):
                self.tree.selection_add(branch)
                self.tree.see(branch)
                break

    "======================================================"
    "=============== widget functions ====================="
    "======================================================"

    def open_nexus_treeview(self):
        filename, folderpath = self.get_filepath()
        logger.info(f"Opening nexus viewer for filename: {filename}")
        if filename:
            from .. import create_nexus_viewer
            create_nexus_viewer(filename, parent=self.root, config=self.config)

    def open_nexus_plot(self):
        filename, folderpath = self.get_filepath()
        logger.info(f"Opening nexus plot viewer for filename: {filename}")
        if filename:
            from ..apps.nexus import create_nexus_plotter
            create_nexus_plotter(filename, parent=self.root, config=self.config)

    def open_nexus_image(self):
        filename, folderpath = self.get_filepath()
        logger.info(f"Opening nexus image viewer for filename: {filename}")
        if filename:
            from ..apps.nexus import create_nexus_image_plotter
            create_nexus_image_plotter(filename, parent=self.root, config=self.config)

    "======================================================"
    "================= misc functions ====================="
    "======================================================"

    def fun_search(self, event=None):
        self.tree.selection_remove(self.tree.selection())
        query = self.search_box.get()
        match_case = self.search_matchcase.get()
        whole_word = self.search_wholeword.get()
        query = query if match_case else query.lower()

        for branch in self.tree.get_children():  # folders
            # folder = self.tree.item(branch)['text']
            for leaf in self.tree.get_children(branch):  # files
                item = self.tree.item(leaf)
                if len(item['values']) < 3:
                    continue
                file = item['text']
                value = item['values'][2]
                test = f"{file} {value}"
                test = test if match_case else test.lower()
                test = test.split() if whole_word else test
                if query in test:
                    self.tree.selection_add(leaf)
                    self.tree.see(leaf)


class FolderScanSelector(_ScanSelector):
    """Frame with TreeView for selection of folders"""

    def __init__(self, root: tk.Misc, initial_directory: str | None = None,
                 config: dict | None = None):
        logger.info('Creating FolderScanSelector')
        super().__init__(root, config)

        # Build widgets
        # self.ini_folderpath()
        self.ini_file_select()
        self.pack_treeview()  # pack treeview after ini_file_select
        self.tree.bind("<Double-1>", self.on_double_click)
        # self.tree.bind('<KeyPress>', self.on_key_press)
        self.tree.bind("<Button-3>", self.right_click_menu())

        # Populate
        if initial_directory:
            # run after mainloop starts to avoid thread issues
            root.after(0, self.add_folder, initial_directory)
        root.after(100, self.poll_files, None)

    "======================================================"
    "================= init functions ====================="
    "======================================================"

    def ini_folderpath(self):
        frm = ttk.Frame(self.root)
        frm.pack(side=tk.TOP, fill=tk.X)

        ttk.Button(frm, text='Add Folder', command=self.browse_folder).pack(side=tk.LEFT)
        ttk.Button(frm, text='Search', command=self.search_options).pack(side=tk.RIGHT)

    def ini_file_select(self):
        frm = ttk.Frame(self.root)
        frm.pack(side=tk.TOP, fill=tk.X)

        var = ttk.Entry(frm, textvariable=self.select_box, width=12)
        var.pack(side=tk.LEFT)
        var.bind("<Return>", self.select_from_box)
        var.bind('<KP_Enter>', self.select_from_box)
        ttk.Button(frm, text='-', command=self.select_box_decrease, width=2).pack(side=tk.LEFT)
        ttk.Button(frm, text='+', command=self.select_box_increase, width=2).pack(side=tk.LEFT)

        ttk.Button(frm, text='Search', command=self.search_options).pack(side=tk.RIGHT)
        ttk.Button(frm, text='Add Folder', command=self.browse_folder).pack(side=tk.RIGHT)

    "======================================================"
    "=============== populate functions ==================="
    "======================================================"

    def add_folder(self, folder_path: str):
        iid = self._add_row("", name=os.path.basename(folder_path), filepath=folder_path)
        files = list_files(folder_path, self.extension.get())
        self.populate_files(iid, *files)

    def update_files(self):
        """Check folders in the tree for new files"""
        for branch in self.tree.get_children():
            folder = self.tree.set(branch, 'filepath')
            files = list_files(folder, self.extension.get())
            for leaf in self.tree.get_children(branch):  # files
                if not self.tree.winfo_exists():
                    return
                file = self.tree.set(leaf, 'filepath')
                if file in files:
                    files.remove(file)

            logger.info(f"Updating {len(files)} in '{os.path.basename(folder)}'")
            logger.debug(f"update_files: Current thread: {current_thread()}, in process pid: {os.getpid()}")
            for file in reversed(files):
                if not self.tree.winfo_exists():
                    return
                iid = self._add_file(branch, file)
                if self.read_datasets.get():
                    self._add_data(iid)

    "======================================================"
    "============= navigation functions ==================="
    "======================================================"

    def browse_folder(self):
        filename, foldername = self.get_filepath()
        folder_directory = select_folder(initial_directory=foldername, parent=self.root)
        if folder_directory:
            self.add_folder(folder_directory)

    def on_select(self, event=None):
        if not self.tree.focus():
            return
        print(self.get_filepath())

    def on_double_click(self, event=None):
        """Action on double click of file"""
        if not self.tree.focus():
            return
        # iid = self.tree.focus()
        # item = self.tree.item(iid)
        self.open_nexus_treeview()

    "======================================================"
    "================= button functions ==================="
    "======================================================"

    def search_options(self):
        filename, foldername = self.get_filepath()

        top = self.root.winfo_toplevel()
        window, fun_close = create_hover(top)
        widget = FindScans(window, foldername, self.config, filename, close_fun=fun_close)

        scan_numbers = widget.wait_for_number()
        if scan_numbers:
            self.tree.selection_remove(self.tree.selection())
            for iid in self.tree.get_children():  # folders
                for scan_iid in self.tree.get_children(iid):
                    scan_number = self.tree.item(scan_iid)['text']
                    if int(scan_number) in scan_numbers:
                        self.tree.selection_add(scan_iid)
                        self.tree.see(scan_iid)
                        break


class ScanViewer(_ScanSelector):
    """Frame with TreeView for selection of scans"""

    def __init__(self, root: tk.Misc, *scan_files: str, config: dict | None = None, button_name: str = 'Close'):
        logger.info('Creating ScanViewer')
        super().__init__(root, config)
        self.pack_treeview()
        self.file_list = scan_files
        self.output_files = []

        self.tree.bind("<Button-3>", self.right_click_menu())

        ttk.Button(self.root, text=button_name, command=self.select_scans).pack(side=tk.TOP, fill=tk.X, expand=tk.YES)

        self.populate_files("", *scan_files)

    def update_metadata(self, event=None):
        """Update dataset values column for hdf files under open folders"""

        def fn():
            for file in self.tree.get_children():  # files
                if not self.tree.winfo_exists():
                    return
                self._add_data(file)

        th = Thread(target=fn)
        th.daemon = True  # runs thread in the background, outside mainloop, allowing python to close
        th.start()

    def select_scans(self):
        self.output_files = [
            self.tree.set(iid, column='filepath')
            for iid in self.tree.selection()
        ]
        self.root.destroy()

    def show(self):
        self.root.wait_window()
        return self.output_files





"""
widget for performing peak fitting on a list of scans
"""

import tkinter as tk
from tkinter import ttk, messagebox
import matplotlib.pyplot as plt

import hdfmap
import numpy as np

from mmg_toolbox import Experiment
from mmg_toolbox.plotting.matplotlib import generate_subplots, set_span_bounds
from mmg_toolbox.utils.env_functions import get_processing_directory
from mmg_toolbox.utils.fitting import (multipeakfit, FitResults, PEAK_MODELS, BACKGROUND_MODELS,
                                       Model, find_peaks, find_peaks_str)
from ..misc.logging import create_logger
from ..misc.config import get_config, C
from ..misc.functions import show_error
from .treeview import CanvasTreeview
from ..widgets.simple_plot import SimplePlot

logger = create_logger(__file__)

FIT_PARAMETERS = ['amplitude', 'center', 'fwhm', 'background']


class ScanFitModel:
    def __init__(self, scan_no: int, filepath: str, metadata: float, title: str, label: str):
        self.scan_no = scan_no
        self.filepath = filepath
        self.metadata = metadata
        self.x_data = np.array([])
        self.y_data = np.array([])
        self.use_dataset = True
        self.n_peaks: int | None = None
        self.peak_model = 'Gaussian'
        self.background_model = 'Slope'
        self.power: float | None = None
        self.distance = 10
        self.mask: np.ndarray | None = None
        self.title = title
        self.label = label
        self.model: Model | None = None
        self.result: FitResults | None = None

    def fit(self) -> FitResults:
        self.result = multipeakfit(
            xvals=self.x_data,
            yvals=self.y_data,
            yerrors=None,
            npeaks=self.n_peaks,
            min_peak_power=self.power,
            peak_distance_idx=self.distance,
            model=self.peak_model,
            background=self.background_model,
            initial_parameters=None,
            fix_parameters=None,
            method='leastsq',
        )
        return self.result


class ScanPeakTreeview(CanvasTreeview):
    """Treeview object for peak details of scans"""
    def __init__(self, root: tk.Misc, width: int | None = None, height: int | None = None):
        columns = [
            ('#0', 'Scan number', 100, False, None),
            ('metadata', 'Metadata', 100, False, None),
            ('model', 'Model', 200, False, None),
            ('filepath', 'Filepath', 0, False, None),
        ]
        super().__init__(root, *columns, width=width, height=height)

    def populate(self, *scan_details: ScanFitModel):
        self.delete()
        for model in scan_details:
            metadata_str = f"{model.metadata:.5}"
            models = f"{model.n_peaks} {model.peak_model} + {model.background_model}"
            values = (metadata_str, models, model.filepath)
            self.tree.insert("", tk.END, text=str(model.scan_no), values=values)

    def get_current_filepath(self):
        iid = next(iter(self.tree.selection()))
        return self.tree.set(iid, 'filepath')

    def first_scan_number(self) -> int:
        iid = self.first_item()
        return int(self.tree.item(iid, 'text'))

    def first_filepath(self):
        iid = self.first_item()
        return self.tree.set(iid, 'filepath')


class PeakFitAnalysis:
    """Frame to perform peak fitting on a set of scans"""

    def __init__(self, root: tk.Misc, config: dict | None = None, exp_directory: str | None = None,
                 proc_directory: str | None = None, scan_numbers: list[int] | None = None,
                 metadata: str | None = None, x_axis: str | None = None, y_axis: str | None = None):
        logger.info('Creating PeakFitAnalysis')
        self.root = root
        self.config = config or get_config()
        self.exp_directory = exp_directory or self.config.get(C.current_dir, '')
        self.proc_directory = proc_directory or self.config.get(C.current_proc, get_processing_directory(self.exp_directory))

        # self.exp_folder = tk.StringVar(root, exp_directory)
        # self.proc_folder = tk.StringVar(root, proc_directory)
        # self.output_file = tk.StringVar(root, proc_directory + '/file.py')
        self.x_axis = tk.StringVar(self.root, 'axes' if x_axis is None else x_axis)
        self.y_axis = tk.StringVar(self.root, 'signal' if y_axis is None else y_axis)
        self.metadata_name = tk.StringVar(self.root, '' if metadata is None else metadata)
        self.all_n_peaks = tk.IntVar(self.root, 1)
        self.all_peak_power = tk.DoubleVar(self.root, 1)
        self.all_peak_distance = tk.IntVar(self.root, 10)
        self.all_model = tk.StringVar(self.root, 'Gaussian')
        self.all_background = tk.StringVar(self.root, 'Slope')
        self.scan_use_scan = tk.BooleanVar(self.root, True)
        self.scan_n_peaks = tk.IntVar(self.root, 1)
        self.scan_peak_power = tk.DoubleVar(self.root, 1)
        self.scan_use_peaks = tk.BooleanVar(self.root, True)
        self.scan_use_power = tk.BooleanVar(self.root, True)
        self.scan_peak_distance = tk.IntVar(self.root, 10)
        self.scan_model = tk.StringVar(self.root, 'Gaussian')
        self.scan_background = tk.StringVar(self.root, 'Slope')
        self.scan_title = tk.StringVar(self.root, '')
        self.scan_label = tk.StringVar(self.root, '')
        self.plot_option = tk.StringVar(self.root, FIT_PARAMETERS[0])
        self.map: hdfmap.NexusMap | None = None
        self.fit_models: list[ScanFitModel] = []

        # ---Top section---
        self.ini_top_section(self.root)

        # ---Middle section---
        middle = ttk.Frame(self.root)
        middle.pack(side=tk.TOP, fill=tk.BOTH, expand=tk.YES, padx=4, pady=4)

        # Left side - scan selection
        left = ttk.LabelFrame(middle, text='Scan Numbers')
        left.pack(side=tk.LEFT, fill=tk.Y, padx=2, pady=2)

        self.scans = ScanPeakTreeview(left)
        self.scans.bind_select(self.select_scan)

        # Right side
        right = ttk.Frame(middle)
        right.pack(side=tk.LEFT, fill=tk.Y, padx=2, pady=2)

        # Plot
        frm = ttk.Frame(right)
        frm.pack(side=tk.TOP, fill=tk.BOTH, padx=2, pady=2)

        self.plot = SimplePlot(frm, [], [], x_axis, y_axis, config=self.config)
        self.data_line, = self.plot.plot([], [], 'bo-', label='Data')
        self.mask_line, = self.plot.plot([], [], '.', label='mask')
        self.fit_line, = self.plot.plot([], [], 'r-', label='Fit')
        self.pl_span = self.plot.ax1.axvspan(0, 0, alpha=0.2)

        # Fit parameters
        frm = ttk.Frame(right)
        frm.pack(side=tk.TOP, fill=tk.BOTH, padx=2, pady=2)

        self.ini_parameters(frm)

        # ---Bottom---
        bottom = ttk.Frame(self.root)
        bottom.pack(side=tk.TOP, expand=tk.YES, pady=8, padx=4)
        ttk.Button(bottom, text='Plot All', command=self.fit_plots).pack(side=tk.LEFT)
        ttk.Button(bottom, text='Plot Fit results', command=self.fit_all).pack(side=tk.LEFT)
        options = ['All'] + FIT_PARAMETERS
        ttk.Combobox(bottom, textvariable=self.plot_option, values=options).pack(side=tk.LEFT, padx=2)

        # ---Start---
        if scan_numbers:
            self.add_scans(*scan_numbers)
            self.scans.tree.selection_set(self.scans.first_item())

    def ini_top_section(self, root: tk.Misc):
        # top = ttk.LabelFrame(root, text='Folders')
        # top.pack(side=tk.TOP, fill=tk.BOTH, expand=tk.YES, padx=4, pady=4)
        #
        # frm = ttk.Frame(top)
        # frm.pack(side=tk.TOP, fill=tk.X, expand=tk.YES, padx=4)
        # ttk.Label(frm, text='Data Dir:', width=15).pack(side=tk.LEFT, padx=4)
        # ttk.Entry(frm, textvariable=self.exp_folder, width=60).pack(side=tk.LEFT)
        # ttk.Button(frm, text='Browse', command=self.browse_datadir).pack(side=tk.LEFT)
        #
        # frm = ttk.Frame(top)
        # frm.pack(side=tk.TOP, fill=tk.X, expand=tk.YES, padx=4)
        # ttk.Label(frm, text='Analysis Dir:', width=15).pack(side=tk.LEFT, padx=4)
        # ttk.Entry(frm, textvariable=self.proc_folder, width=60).pack(side=tk.LEFT)
        # ttk.Button(frm, text='Browse', command=self.browse_analysis).pack(side=tk.LEFT)

        # Axis + Metadata selection
        top = ttk.LabelFrame(root, text='Axes')
        top.pack(side=tk.TOP, fill=tk.BOTH, expand=tk.YES, padx=4, pady=4)

        frm = ttk.Frame(top)
        frm.pack(side=tk.TOP, fill=tk.X, expand=tk.YES, padx=4)
        ttk.Label(frm, text='X:', width=2).pack(side=tk.LEFT, padx=2)
        var = ttk.Entry(frm, textvariable=self.x_axis, width=20)
        var.pack(side=tk.LEFT)
        var.bind("<Return>", self.plot_scan)
        var.bind("<KP_Enter>", self.plot_scan)
        ttk.Button(frm, text=':', command=self.browse_x_axis, width=1, padding=0).pack(side=tk.LEFT)

        ttk.Label(frm, text='Y:', width=2).pack(side=tk.LEFT, padx=4)
        var = ttk.Entry(frm, textvariable=self.y_axis, width=20)
        var.pack(side=tk.LEFT)
        var.bind("<Return>", self.plot_scan)
        var.bind("<KP_Enter>", self.plot_scan)
        ttk.Button(frm, text=':', command=self.browse_y_axis, width=1, padding=0).pack(side=tk.LEFT)

        ttk.Label(frm, text='Metadata:', width=15).pack(side=tk.LEFT, padx=4)
        var = ttk.Entry(frm, textvariable=self.metadata_name, width=20)
        var.pack(side=tk.LEFT)
        var.bind("<Return>", self.update_scans)
        var.bind("<KP_Enter>", self.update_scans)
        ttk.Button(frm, text=':', command=self.browse_metadata, width=1, padding=0).pack(side=tk.LEFT)
        ttk.Button(frm, text='Update', command=self.update_scans, padding=1).pack(side=tk.RIGHT, padx=5)

        # Model selection
        frm = ttk.LabelFrame(root, text='Fit Options')
        frm.pack(side=tk.TOP, fill=tk.BOTH, expand=tk.YES, padx=4, pady=4)

        ttk.Label(frm, text='N Peaks:', width=8).pack(side=tk.LEFT, padx=2)
        ttk.Entry(frm, textvariable=self.all_n_peaks, width=3).pack(side=tk.LEFT)

        ttk.Label(frm, text='Power:', width=8).pack(side=tk.LEFT, padx=2)
        ttk.Button(frm, text='?', command=self.help_power, width=2, padding=0).pack(side=tk.LEFT)
        ttk.Entry(frm, textvariable=self.all_peak_power, width=3).pack(side=tk.LEFT)

        ttk.Label(frm, text='Distance:', width=8).pack(side=tk.LEFT, padx=2)
        ttk.Button(frm, text='?', command=self.help_peak_distance, width=2, padding=0).pack(side=tk.LEFT)
        ttk.Entry(frm, textvariable=self.all_peak_distance, width=3).pack(side=tk.LEFT)

        ttk.Label(frm, text='Model:', width=12).pack(side=tk.LEFT, padx=4)
        ttk.Combobox(frm, textvariable=self.all_model, values=list(PEAK_MODELS)).pack(side=tk.LEFT, padx=2)

        ttk.Label(frm, text='Background:', width=12).pack(side=tk.LEFT, padx=4)
        ttk.Combobox(frm, textvariable=self.all_background, values=list(BACKGROUND_MODELS)).pack(side=tk.LEFT, padx=2)

    def ini_parameters(self, root: tk.Misc):
        frm = ttk.Frame(root)
        frm.pack(side=tk.TOP, fill=tk.X, expand=tk.YES, padx=4, pady=2)
        left = ttk.Frame(frm)
        left.pack(side=tk.LEFT)
        line = ttk.Frame(left)
        line.pack(side=tk.TOP)
        ttk.Checkbutton(line, variable=self.scan_use_scan, command=self.update_model).pack(side=tk.LEFT)
        ttk.Label(line, textvariable=self.scan_title).pack(side=tk.LEFT, fill=tk.X, padx=5, pady=2)
        ttk.Label(left, textvariable=self.scan_label, width=20).pack(side=tk.TOP, fill=tk.X, padx=5, pady=2)
        right = ttk.Frame(frm)
        right.pack(side=tk.LEFT)
        ttk.Button(right, text='Fit', command=self.perform_fit).pack(side=tk.LEFT, fill=tk.Y, padx=5, pady=2)
        ttk.Button(right, text='Results', command=self.display_results).pack(side=tk.LEFT, fill=tk.Y, padx=5, pady=2)

        frm = ttk.Frame(root)
        frm.pack(side=tk.TOP, fill=tk.X, expand=tk.YES, padx=4, pady=2)
        opt = ttk.Frame(frm, relief=tk.GROOVE)
        opt.pack(side=tk.LEFT, padx=5, pady=4)
        ttk.Label(opt, text='N Peaks:', width=10).pack(side=tk.LEFT, padx=2)
        ttk.Entry(opt, textvariable=self.scan_n_peaks, width=3).pack(side=tk.LEFT)
        ttk.Checkbutton(opt, variable=self.scan_use_peaks, command=self.update_model, padding=0).pack(side=tk.LEFT, padx=2)

        opt = ttk.Frame(frm, relief=tk.GROOVE)
        opt.pack(side=tk.LEFT, padx=5, pady=4)
        ttk.Label(opt, text='Power:', width=10).pack(side=tk.LEFT, padx=2)
        ttk.Entry(opt, textvariable=self.scan_peak_power, width=3).pack(side=tk.LEFT)
        ttk.Checkbutton(opt, variable=self.scan_use_power, command=self.update_model, padding=0).pack(side=tk.LEFT, padx=2)

        opt = ttk.Frame(frm, relief=tk.GROOVE)
        opt.pack(side=tk.LEFT, padx=5, pady=4)
        ttk.Label(opt, text='Distance:', width=10).pack(side=tk.LEFT, padx=2)
        ttk.Entry(opt, textvariable=self.scan_peak_distance, width=3).pack(side=tk.LEFT)

        frm = ttk.Frame(root)
        frm.pack(side=tk.TOP, fill=tk.X, expand=tk.YES, padx=4, pady=2)
        ttk.Label(frm, text='Model:').pack(side=tk.LEFT, padx=4)
        ttk.Combobox(frm, textvariable=self.scan_model, values=list(PEAK_MODELS), width=10).pack(side=tk.LEFT, padx=2)

        ttk.Label(frm, text='Background:').pack(side=tk.LEFT, padx=4)
        ttk.Combobox(frm, textvariable=self.scan_background, values=list(BACKGROUND_MODELS),
                     width=10).pack(side=tk.LEFT, padx=2)

        frm = ttk.Frame(root)
        frm.pack(side=tk.TOP, padx=4, pady=2)
        ttk.Button(frm, text='Select Region', command=self.select_region).pack(side=tk.LEFT, padx=2)
        ttk.Button(frm, text='Mask Region', command=self.select_mask).pack(side=tk.LEFT, padx=2)
        ttk.Button(frm, text='Reset', command=self.reset_mask).pack(side=tk.LEFT, padx=2)

        # frm = ttk.Frame(root)
        # frm.pack(side=tk.TOP, fill=tk.X, expand=tk.YES, padx=4)
        # parameters
        # TODO: add parameters

    def browse_x_axis(self):
        from ..apps.namespace_select import create_scannable_selector
        names = create_scannable_selector(self.map)
        if names:
            self.x_axis.set(', '.join(name for name in names))
            self.plot_scan()

    def browse_y_axis(self):
        from ..apps.namespace_select import create_scannable_selector
        names = create_scannable_selector(self.map)
        if names:
            self.y_axis.set(', '.join(name for name in names))
            self.plot_scan()

    def browse_metadata(self):
        from ..apps.namespace_select import create_metadata_selector
        paths = create_metadata_selector(self.map)
        if paths:
            self.metadata_name.set(', '.join(path for path in paths))
            self.update_scans()

    def help_power(self):
        messagebox.showinfo(
            title='Peak Power',
            message=(
                'A peak must achieve this ratio of signal / background to count as a peak.\n' +
                'Put 0 to allow all peaks.'
            ),
            parent=self.root,
        )

    def help_peak_distance(self):
        messagebox.showinfo(
            title='Peak Distance',
            message='Multiple peaks must be separated by this distance, in units of x-elements',
            parent=self.root,
        )

    #######################################################
    ################## methods ############################
    #######################################################

    def get_experiment(self):
        return Experiment(self.exp_directory, instrument=self.config.get('beamline', None))

    def get_scan_files(self, *scan_numbers: int) -> list[str]:
        try:
            exp = self.get_experiment()
            scan_files = [exp.get_scan_filename(n) for n in scan_numbers]
            self.map = hdfmap.create_nexus_map(scan_files[0])
        except Exception as e:
            show_error(e, self.root, raise_exception=False)
            raise e
        return scan_files

    def get_metadata(self, *scan_files: str) -> np.ndarray:
        name = self.metadata_name.get().split(',')[0]
        metadata = np.zeros(len(scan_files))
        if not name:
            return metadata
        for n, filename in enumerate(scan_files):
            with hdfmap.load_hdf(filename) as hdf:
                metadata[n] = self.map.eval(hdf, name)
        return metadata

    def add_scans(self, *scan_numbers: int):
        scan_files = self.get_scan_files(*scan_numbers)
        self.fit_models = [
            ScanFitModel(
                scan_no=n,
                filepath=f,
                metadata=0,
                title=str(n),
                label=f
            ) for n, f in zip(scan_numbers, scan_files)
        ]
        self.update_scans()

    def update_scans(self, event=None):
        metadata = self.get_metadata(*(model.filepath for model in self.fit_models))
        for n, model in enumerate(self.fit_models):
            model.metadata = metadata[n]
            model.peak_model = self.all_model.get()
            model.background_model = self.all_background.get()
            model.n_peaks = self.all_n_peaks.get()
            model.power = self.all_peak_power.get()
            model.distance = self.all_peak_distance.get()
        self.scans.populate(*self.fit_models)

    def get_scan_xy_data(self, filename: str | None = None) -> tuple[np.ndarray, np.ndarray]:
        if filename is None:
            filename = self.scans.get_current_filepath()
        x_axis = self.x_axis.get()
        y_axis = self.y_axis.get()
        with self.map.load_hdf(filename) as hdf:
            x_data = self.map.eval(hdf, x_axis, np.arange(self.map.scannables_length()))
            y_data = self.map.eval(hdf, y_axis, np.ones_like(x_data))
        return x_data, y_data

    def select_scan(self, event=None):
        index = self.scans.get_index()
        model = self.fit_models[index]
        self.scan_use_scan.set(model.use_dataset)
        self.scan_n_peaks.set(0 if model.n_peaks is None else model.n_peaks)
        self.scan_use_peaks.set(False if model.n_peaks is None else True)
        self.scan_model.set(model.peak_model)
        self.scan_background.set(model.background_model)
        self.scan_peak_power.set(0 if model.power is None else model.power)
        self.scan_use_power.set(False if model.power is None else True)
        self.scan_peak_distance.set(model.distance)
        self.scan_title.set(model.title)
        self.scan_label.set(model.label)
        self.perform_fit()

    def update_model(self, index: int | None = None) -> ScanFitModel:
        if index is None:
            index = self.scans.get_index()
        model = self.fit_models[index]
        x_data, y_data = self.get_scan_xy_data(model.filepath)
        mask = self.fit_models[self.scans.get_index()].mask
        if mask is not None:
            x_data = x_data[mask]
            y_data = y_data[mask]
        model.x_data = x_data
        model.y_data = y_data
        model.use_dataset = self.scan_use_scan.get()
        model.n_peaks = self.scan_n_peaks.get() if self.scan_use_peaks.get() else None
        model.peak_model = self.scan_model.get()
        model.background_model = self.scan_background.get()
        model.power = self.scan_peak_power.get() if self.scan_use_power.get() else None
        model.distance = self.scan_peak_distance.get()
        return model

    def plot_scan(self, event=None):
        model = self.fit_models[self.scans.get_index()]
        x_data, y_data = self.get_scan_xy_data(model.filepath)
        if model.mask is not None:
            x_mask = x_data[model.mask]
            y_mask = y_data[model.mask]
            x_data = x_data[~model.mask]
            y_data = y_data[~model.mask]
        else:
            x_mask, y_mask = [], []

        if model.result is not None:
            x_fit, y_fit = model.result.fit_data(x_data)
        else:
            x_fit, y_fit = [], []

        x_label, y_label = self.map.generate_ids(self.x_axis.get(), self.y_axis.get())

        self.data_line.set_xdata(x_data)
        self.data_line.set_ydata(y_data)
        self.mask_line.set_xdata(x_mask)
        self.mask_line.set_ydata(y_mask)
        self.fit_line.set_xdata(x_fit)
        self.fit_line.set_ydata(y_fit)
        self.plot.ax1.set_xlabel(x_label)
        self.plot.ax1.set_ylabel(y_label)
        set_span_bounds(self.pl_span, x_data[0], x_data[0], y_data[0], y_data[0])
        # self.pl_span.set_bounds(x_data[0], y_data[0], 0, 0)
        self.plot.update_axes()

    def perform_fit(self):
        model = self.update_model()
        model.fit()
        self.plot_scan()

    def display_results(self):
        from ..apps.edit_text import EditText
        model = self.update_model()
        if model.result is None:
            return
        peak_str = find_peaks_str(model.x_data, model.y_data,
                                  min_peak_power=model.power, peak_distance_idx=model.distance)

        meta_str = f"{self.metadata_name.get()} = {model.metadata:.5g}"
        out = f"{model.title}\n{model.label}\n{meta_str}\n\n"
        # out += f"Power = {peak_ratio(model.y_data)}\n\n"
        out += peak_str + '\n\n'
        out += str(model.result)
        EditText(out, parent=self.root, title=model.label)

    def scans_title(self):
        scan_numbers = [model.scan_no for model in self.fit_models]
        exp = self.get_experiment()
        return exp.generate_scans_title(*scan_numbers, hdf_map=self.map)

    def fit_all(self):
        option = self.plot_option.get()
        if option == 'All':
            self._plot_all_results()
            return

        metadata, value, error = [], [], []
        for n in range(len(self.fit_models)):
            model = self.update_model(n)
            if model.use_dataset:
                result = model.fit()
                metadata.append(model.metadata)
                val, err = result.get_value(option)
                value.append(val)
                error.append(err)

        x_label = self.metadata_name.get()
        y_label = option.capitalize()

        fig, ax = plt.subplots()
        ax.errorbar(metadata, value, error, fmt='-o')
        ax.set_xlabel(x_label)
        ax.set_ylabel(y_label)
        ax.set_title(self.scans_title())
        plt.show()

    def _plot_all_results(self):
        dpi = self.config.get(C.plot_dpi)
        fig, axes = plt.subplots(2, 2, figsize=[12, 12], dpi=dpi)
        fig.suptitle(self.scans_title())
        axes = axes.flatten()
        x_label = self.metadata_name.get()
        metadata = [model.metadata for model in self.fit_models]
        results = [model.fit() for model in self.fit_models]
        for ax, option in zip(axes, FIT_PARAMETERS):
            values, errors = zip(*[result.get_value(option) for result in results])
            ax.errorbar(metadata, values, errors, fmt='-o', label=option)
            ax.set_xlabel(x_label)
            ax.set_ylabel(option.capitalize())
        plt.show()

    def fit_plots(self):
        x_label, y_label, name = self.map.generate_ids(self.x_axis.get(), self.y_axis.get(), self.metadata_name.get())

        fig_axes = generate_subplots(len(self.fit_models))
        for n, (fig, axes) in enumerate(fig_axes):
            model = self.update_model(n)
            x_data, y_data = self.get_scan_xy_data(model.filepath)
            if model.mask is not None:
                x_mask = x_data[model.mask]
                y_mask = y_data[model.mask]
                x_data = x_data[~model.mask]
                y_data = y_data[~model.mask]
            else:
                x_mask, y_mask = [], []

            if model.use_dataset:
                result = model.fit()
                x_fit, y_fit = result.fit_data(x_data)
            else:
                x_fit, y_fit = [], []

            title = f"{model.title}\n{name} = {model.metadata:.5g}"

            axes.plot(x_mask, y_mask, 'b.', label=None)
            axes.plot(x_data, y_data, 'b-o', label='Data')
            axes.plot(x_fit, y_fit, 'r-', label='Fit')
            axes.set_xlabel(x_label)
            axes.set_ylabel(y_label)
            axes.legend()
            axes.set_title(title)
        plt.show()

    "------------------------------------------------------------------------"
    "---------------------------Button Functions-----------------------------"
    "------------------------------------------------------------------------"

    def chk_y(self):
        """Apply functions to y-axis"""
        # old_log, old_smooth, old_diff = self.ydata_checkboxes
        # log = self.ylog.get()
        # smooth = self.ysmooth.get()
        # diff = self.ydiff.get()
        # # update original data
        # if sum(self.ydata_checkboxes) == 0:
        #     xdata, ydata, yerror = self.gen_data()
        #     self.ydata_original = ydata
        # data = 1.0 * self.ydata_original
        # if log:
        #     data = np.log10(data)
        # if diff:
        #     data = np.gradient(data)
        # if smooth:
        #     data = i16_peakfit.functions.conv_gauss(data, len(data) // 4, 0.1)
        # self.ydata_checkboxes = [log, smooth, diff]
        # self.txt_y.delete("1.0", tk.END)
        # self.txt_y.insert("1.0", str(list(data)))
        # self.plot_data()

    def but_find_peaks(self):
        """Button Find Peaks"""
        x_data, y_data = self.get_scan_xy_data()
        mask = self.fit_models[self.scans.get_index()].mask
        if mask is not None:
            x_data = x_data[mask]
            y_data = y_data[mask]
        model = self.update_model()

        # Run find peaks, create message
        idx, pwr = find_peaks(y_data, None, min_peak_power=model.power, peak_distance_idx=model.distance)
        s = "Found %d peaks\n" % len(idx)
        s += "  Index | Position | Power\n"
        for _idx, _pwr in zip(idx, pwr):
            s += f"  {_idx:5} | {x_data[_idx]:8.4g} | {_pwr:5.2f}\n"
        # self.message.set('Found %d peaks' % len(idx))
        # self.txt_res.delete('1.0', tk.END)
        # self.txt_res.insert('1.0', s)
        print(s) #TODO: Do something with this

    def select_peaks(self):
        """Button Select Peaks"""

        def get_mouseposition(event):
            print(event.x, event.y, event.xdata, event.ydata, event.inaxes)
            self.plot.fig.canvas.mpl_disconnect(press)
            # self.root.unbind("<Button-1>")
            self.plot.root.config(cursor="arrow")

            if event.inaxes:
                print(f"Position selected: ({event.xdata:.4}, {event.ydata:.4})")
                # self.message.set('Position selected: (%.4g, %.4g)' % (event.xdata, event.ydata))
                # Add peak
                # self.add_peak('Gaussian', event.xdata, event.ydata)

            else:
                # self.message.set('No position selected')
                print('No position selected')

        # self.message.set('Click on Figure to add a peak')
        press = self.plot.fig.canvas.mpl_connect('button_press_event', get_mouseposition)
        # self.root.bind("<Button-1>", get_mouseposition)
        self.plot.root.config(cursor="crosshair")

    def select_region(self):
        """Button click to select"""

        x_data, ydata = self.get_scan_xy_data()
        model = self.fit_models[self.scans.get_index()]
        mask = model.mask
        if mask is None or len(ydata) != len(mask):
            mask = np.ones(x_data.shape, dtype=bool)

        xval = [np.min(x_data)]
        ipress = [False]

        def disconnect():
            self.plot_scan()
            self.plot.fig.canvas.mpl_disconnect(press)
            self.plot.fig.canvas.mpl_disconnect(move)
            self.plot.fig.canvas.mpl_disconnect(release)
            self.plot.root.config(cursor="arrow")

        def mouse_press(event):
            if event.inaxes:
                xval[0] = event.xdata
                ipress[0] = True
            else:
                disconnect()

        def mouse_move(event):
            if event.inaxes and ipress[0]:
                x_max = event.xdata
                ax_ymin, ax_ymax = self.plot.ax1.get_ylim()
                # span = [[xval[0], ax_ymin], [xval[0], ax_ymax], [x_max, ax_ymax], [x_max, ax_ymin], [xval[0], ax_ymin]]
                # self.pl_span.set_xy(span)
                set_span_bounds(self.pl_span, xval[0], x_max, ax_ymin, ax_ymax)
                # self.pl_span.set_bounds(xval[0], ax_ymin, x_max - xval[0], ax_ymax - ax_ymin)
                self.plot.update_axes()

        def mouse_release(event):
            x_min = min(xval[0], event.xdata)
            x_max = max(xval[0], event.xdata)
            new_selection = (x_data > x_min) * (x_data < x_max)
            if np.any(mask[new_selection]):
                mask[new_selection] = False
            else:
                mask[~new_selection] = True
            model.mask = mask
            disconnect()

        self.plot_scan()
        # self.message.set('Click & Drag region to select')
        press = self.plot.fig.canvas.mpl_connect('button_press_event', mouse_press)
        move = self.plot.fig.canvas.mpl_connect('motion_notify_event', mouse_move)
        release = self.plot.fig.canvas.mpl_connect('button_release_event', mouse_release)
        # self.root.bind("<Button-1>", get_mouseposition)
        self.plot.root.config(cursor="crosshair")

    def select_mask(self):
        """Button click to mask"""

        xdata, ydata = self.get_scan_xy_data()
        mask = self.fit_models[self.scans.get_index()].mask
        if mask is None or len(ydata) != len(mask):
            mask = np.zeros(xdata.shape, dtype=bool)

        xval = [np.min(xdata)]
        ipress = [False]

        def disconnect():
            self.plot_scan()
            self.plot.fig.canvas.mpl_disconnect(press)
            self.plot.fig.canvas.mpl_disconnect(move)
            self.plot.fig.canvas.mpl_disconnect(release)
            self.plot.root.config(cursor="arrow")

        def mouse_press(event):
            if event.inaxes:
                xval[0] = event.xdata
                ipress[0] = True
            else:
                disconnect()

        def mouse_move(event):
            if event.inaxes and ipress[0]:
                x_max = event.xdata
                ax_ymin, ax_ymax = self.plot.ax1.get_ylim()
                # span = [[xval[0], ax_ymin], [xval[0], ax_ymax], [x_max, ax_ymax], [x_max, ax_ymin], [xval[0], ax_ymin]]
                # self.pl_span.set_xy(span)
                set_span_bounds(self.pl_span, xval[0], x_max, ax_ymin, ax_ymax)
                # self.pl_span.set_bounds(xval[0], ax_ymin, x_max - xval[0], ax_ymax - ax_ymin)
                self.plot.update_axes()

        def mouse_release(event):
            x_min = min(xval[0], event.xdata)
            x_max = max(xval[0], event.xdata)
            new_selection = (xdata > x_min) * (xdata < x_max)
            mask[new_selection] = True
            disconnect()

        self.plot_scan()
        # self.message.set('Click & Drag region to select')
        press = self.plot.fig.canvas.mpl_connect('button_press_event', mouse_press)
        move = self.plot.fig.canvas.mpl_connect('motion_notify_event', mouse_move)
        release = self.plot.fig.canvas.mpl_connect('button_release_event', mouse_release)
        # self.root.bind("<Button-1>", get_mouseposition)
        self.plot.root.config(cursor="crosshair")

    def reset_mask(self):
        xdata, ydata = self.get_scan_xy_data()
        self.fit_models[self.scans.get_index()].mask = np.zeros(xdata.shape, dtype=bool)
        self.plot_scan()



"""
a tkinter frame with 4 sections:
    NW: File selection treeview
    SW: NeXus metadata viewer
    NE: 2D Line plot
    SE: Image plot
"""
import tkinter as tk
from tkinter import ttk

from hdfmap import create_nexus_map

from mmg_toolbox.utils.env_functions import get_scan_number
from ..misc.logging import create_logger
from ..misc.config import get_config, C
from .scan_selector import FolderScanSelector
from .nexus_details import NexusDetails
from .nexus_plot_and_image import NexusPlotAndImage

logger = create_logger(__file__)


class NexusDataViewer:
    """
    tkinter widget containing scan selector, details,
    line plot and image plot - the main frame in the data viewer.

    widget = NexusDataViewer(root, 'initial/folder', config)


    """
    def __init__(self, root: tk.Misc, initial_folder: str | None = None,
                 config: dict | None = None):
        self.root = root
        self.map = None
        self.config = config or get_config()

        self.root.rowconfigure(0, weight=1)
        self.root.columnconfigure(0, weight=1)
        grid_options = dict(padx=5, pady=5, sticky='nsew')

        window = ttk.Frame(self.root)
        # window.pack(fill=tk.BOTH, expand=tk.YES)
        window.grid(column=0, row=0, **grid_options)
        window.rowconfigure(0, weight=1)
        window.rowconfigure(1, weight=1)
        window.columnconfigure(0, weight=0)
        window.columnconfigure(1, weight=1)

        # TOP-LEFT
        frm = ttk.LabelFrame(window, text='Files')
        frm.grid(column=0, row=0, **grid_options)
        self.selector_widget = FolderScanSelector(frm, initial_directory=initial_folder, config=self.config)
        self.selector_widget.tree.bind("<<TreeviewSelect>>", self.on_file_select)

        # BOTTOM-LEFT
        frm = ttk.LabelFrame(window, text='Details')
        # frm.pack(side=tk.LEFT, fill=tk.Y, expand=tk.YES, padx=2, pady=2)
        frm.grid(column=0, row=1, **grid_options)
        self.detail_widget = NexusDetails(frm, config=self.config)

        # RIGHT-SIDE
        frm = ttk.Frame(window)
        frm.grid(column=1, row=0, rowspan=2, **grid_options)
        self.plot_widget = NexusPlotAndImage(frm, config=self.config)

        # select first file if it exists
        self.root.after(100, self.select_first_file, None)

    def select_first_file(self, _event=None):
        if len(self.selector_widget.tree.get_children()) > 0:
            first_folder = next(iter(self.selector_widget.tree.get_children()))
            if len(self.selector_widget.tree.get_children(first_folder)) > 0:
                first_scan = next(iter(self.selector_widget.tree.get_children(first_folder)))
                self.selector_widget.tree.item(first_folder, open=True)
                self.selector_widget.tree.selection_set(first_scan)

    def on_file_select(self, event=None):
        filename, folder = self.selector_widget.get_filepath()
        filenames = self.selector_widget.get_multi_filepath()
        if len(filenames) == 0:
            return
        self.config[C.current_dir] = folder

        logger.info(f"Updating widgets for file: {filename}")
        # TODO: time and speed up this part
        self.selector_widget.select_box.set(get_scan_number(filename))
        self.map = create_nexus_map(filename)
        self.detail_widget.update_data_from_file(filename, self.map)
        self.plot_widget.update_data_from_files(*filenames, hdf_map=self.map)


"""
Configuration Options
"""
from __future__ import annotations

import os
import json

from mmg_toolbox.utils.env_functions import TMPDIR, get_beamline, get_user, check_file_access
import mmg_toolbox.beamline_metadata.config as config
from .matplotlib import FIGURE_SIZE, FIGURE_DPI, IMAGE_SIZE, DEFAULT_COLORMAP


class C(config.C):
    """Names used in config object"""
    processing_directory = 'processing_directory'
    notebook_directory = 'notebook_directory'
    recent_data_directories = 'recent_data_directories'
    small_screen_height = 'small_screen_height'
    text_size = 'text_size'
    text_size_small = 'text_size_small'
    plot_size = 'plot_size'
    image_size = 'image_size'
    plot_max_percent = 'plot_max_percent'
    plot_dpi = 'plot_dpi'
    plot_title = 'plot_title'
    default_colormap = 'default_colormap'
    current_dir = 'current_dir'
    current_proc = 'current_proc'
    current_nb = 'current_nb'


# config name (saved in TMPDIR)
USER = get_user()
TMPFILE = f'mmg_config_{USER}.json'
CONFIG_FILE = os.path.join(TMPDIR, TMPFILE)
SMALL_SCREEN_HEIGHT = 800  # pixels, reduce size if screen smaller than this
TEXT_WIDTH = 50  # Determines the width of text areas in DataViewer in characters
TEXT_HEIGHT = 25  # Determines height of text area in Dataviewer in lines
TEXT_HEIGHT_SMALL = 10  # TEXT_HEIGHT when screen is small
MAX_PLOT_SCREEN_PERCENTAGE = (75, 25)  # (wid, height) max plot size as % of screen
BEAMLINE_CONFIG = config.BEAMLINE_CONFIG


CONFIG = {
    **config.CONFIG,
    C.conf_file: CONFIG_FILE,
    C.default_directory: os.path.expanduser('~'),
    C.processing_directory: os.path.expanduser('~'),
    C.notebook_directory: os.path.expanduser('~'),
    C.recent_data_directories: [os.path.expanduser('~')],
    C.small_screen_height: SMALL_SCREEN_HEIGHT,
    C.text_size: (TEXT_WIDTH, TEXT_HEIGHT),
    C.text_size_small: (TEXT_WIDTH, TEXT_HEIGHT_SMALL),
    C.plot_size: FIGURE_SIZE,
    C.image_size: IMAGE_SIZE,
    C.plot_max_percent: MAX_PLOT_SCREEN_PERCENTAGE,
    C.plot_dpi: FIGURE_DPI,
    C.plot_title: '{filename}\n{(cmd|scan_command)}',
    C.default_colormap: DEFAULT_COLORMAP,
}

def check_config_filename(config_filename: str | None) -> str:
    """Check config filename is writable, raise OSError if not"""
    if config_filename is None:
        config_filename = CONFIG_FILE
    return check_file_access(config_filename)


def load_config(config_filename: str = CONFIG_FILE) -> dict:
    """Loads a config dict from file, by default from the default location"""
    if os.path.isfile(config_filename):
        with open(config_filename, 'r') as f:
            return json.load(f)
    return {}


def default_config(beamline: str | None = None) -> dict:
    """Returns the default beamline config dict"""
    cfg = CONFIG.copy()
    if beamline is None:
        beamline = get_beamline()
    if beamline in config.BEAMLINE_CONFIG:
        cfg.update(config.BEAMLINE_CONFIG[beamline])
    return cfg


def get_config(config_filename: str | None = None, beamline: str | None = None) -> dict:
    """merge loaded config into default beamline config and return the config dict"""
    config_filename = check_config_filename(config_filename)
    user_config = load_config(config_filename)
    cfg = default_config(beamline)
    if beamline and user_config.get(C.beamline) and beamline != user_config.get(C.beamline):
        # default config overrides user config when changing beamline
        user_config.update(cfg)
        return user_config
    cfg.update(user_config)
    return cfg


def reset_config(cfg: dict) -> None:
    """Reset config dict in place with default values of beamline"""
    beamline = cfg.get(C.beamline, None)
    cfg.clear()
    cfg.update(default_config(beamline))


def save_config(cfg: dict):
    """Save the config dict into the file location referenced in config['config_file']"""
    config_filename = cfg.get(C.conf_file, CONFIG_FILE)
    with open(config_filename, 'w') as f:
        json.dump(cfg, f)
    print('Saved config to {}'.format(config_filename))


def save_config_as(config_filename: str | None = None, **kwargs):
    """Save the config dict into a new file location"""
    cfg = get_config(config_filename)
    cfg.update(kwargs)
    cfg[C.conf_file] = config_filename
    save_config(cfg)




"""
Various tkinter functions
"""

import os
import tkinter
import tkinter as tk
from tkinter import ttk, filedialog, messagebox
from PIL import ImageGrab

from .styles import RootWithStyle, theme_menu


def topmenu(root: RootWithStyle, menu_dict: dict, add_themes=False, add_about=False,
            config: dict | None = None):
    """
    Add a file menu to root
    :param root: tkinter root
    :param menu_dict: {Menu name: {Item name: function}}
    :param add_themes: add themes menu item
    :param add_about: add about menu item
    :param config: add Config menu if config dict is added
    :return: None
    """
    if config is not None:
        menu_dict.update(config_menu(root, config))
    if add_themes and hasattr(root, 'style'):
        menu_dict.update(theme_menu(root.style))
    if add_about:
        menu_dict.update(about_menu(root))

    def add_menu(menu: tk.Menu, **this_menu_dict):
        for name, item in this_menu_dict.items():
            if type(item) is dict:
                new_menu = tk.Menu(menu, tearoff=False)
                add_menu(new_menu, **item)
                menu.add_cascade(label=name, menu=new_menu)
            else:
                menu.add_command(label=name, command=item)

    menubar = tk.Menu(root)
    add_menu(menubar, **menu_dict)
    root.config(menu=menubar)


def about_menu(root: tk.Misc | None = None):
    """About menu items"""
    menu = {
        'Help': {
            'Docs': open_docs,
            'About': lambda: popup_about(root)
        }
    }
    return menu


def config_menu(root: tk.Misc, config: dict) -> dict:
    """Config menu items"""
    from ..apps.config_editor import ConfigEditor
    from .config import reset_config
    menu = {
        'Config': {
            'Edit Config.': lambda: ConfigEditor(root, config),
            'Reset Config.': lambda: reset_config(config),
        }
    }
    return menu


def select_hdf_file(parent):
    """Select HDF file using filedialog"""
    from h5py import is_hdf5
    filename = filedialog.askopenfilename(
        title='Select file to open',
        filetypes=[('NXS file', '.nxs'),
                   ('HDF file', '.h5'), ('HDF file', '.hdf'), ('HDF file', '.hdf5'),
                   ('All files', '.*')],
        parent=parent
    )
    if filename and not is_hdf5(filename):
        messagebox.showwarning(
            title='Incorrect File Type',
            message=f"File: \n{filename}\n can't be read by h5py",
            parent=parent
        )
        filename = None
    return filename


def select_folder(parent, initial_directory: str | None = None):
    """Select folder"""
    foldername = filedialog.askdirectory(
        initialdir=initial_directory,
        title='Select folder...',
        mustexist=True,
        parent=parent,
    )
    return foldername


def check_new_file(parent, filename: str | None = None) -> str | None:
    """Check if filename is new and writable"""
    path, name = os.path.split(filename)
    if not os.access(path, os.W_OK):
        show_error(f"new file cannot be written as path is not writable: '{path}'",
                   parent=parent, raise_exception=False)
        return None

    if os.path.exists(filename):
        if os.access(filename, os.W_OK):
            answer = messagebox.askyesnocancel(
                title='Create File',
                message=f"Overwrite {name}? Yes to overwrite or No to create new file.",
                parent=parent
            )
            if answer is None:
                return None
            if answer:
                return filename
        # amend name
        name, ext = os.path.splitext(name)
        n_tries = 1
        while os.path.exists(filename):
            if n_tries > 100:
                return None
            filename = os.path.join(path, name + str(n_tries) + ext)
            n_tries += 1
    return filename


def open_close_all_tree(treeview, branch="", openstate=True):
    """Open or close all items in ttk.treeview"""
    treeview.item(branch, open=openstate)
    for child in treeview.get_children(branch):
        open_close_all_tree(treeview, child, openstate)  # recursively open children


def treeview_sort_column(treeview: ttk.Treeview, col: str, reverse: bool, sort_col: str | None = None):
    """
    Function to sort columns in ttk.Treeview,
        tree.heading("#0", command=lambda _col="#0": treeview_sort_column(tree, _col, False))
    :param treeview: ttk.Treeview instance
    :param col: str, column specifier for items to sort
    :param reverse: Bool, sort direction
    :param sort_col: str or None, sort alternative column
    :return:
    """
    if sort_col is None:
        sort_col = col
    if col == "#0":
        def item(iid):
            return treeview.item(iid)['text']
    else:
        def item(iid):
            return treeview.set(iid, col)

    items = [(item(iid), iid) for iid in treeview.get_children('')]
    items.sort(reverse=reverse)

    # rearrange items in sorted positions
    for index, (val, k) in enumerate(items):
        treeview.move(k, '', index)
        if treeview.item(k)['text'] == '..':  # keep at top of column
            treeview.move(k, '', 0)

    # reverse sort next time
    treeview.heading(sort_col, command=lambda _col=col: treeview_sort_column(treeview, _col, not reverse, sort_col))


def show_error(message, parent=None, raise_exception=True):
    """Display and raise error"""
    messagebox.showwarning(
        title="HDF File Error",
        message=message,
        parent=parent,
    )
    if raise_exception:
        raise Exception(message)


def open_docs():
    """Open web-browser at docs site"""
    import webbrowser
    webbrowser.open_new_tab("https://diamondlightsource.github.io/mmg_toolbox/")


def popup_about(root: tk.Misc | None = None):
    """Create about message"""
    from mmg_toolbox import version_info, module_info, title
    msg = (
        f"{version_info()}\n\n" +
        "A selection of useful functions and methods for the mmg beamlines at Diamond" +
        "\n\n" +
        f"Module Info:\n{module_info()}\n\n" +
        "By Dan Porter, Diamond Light Source Ltd"
    )
    if root is not None:
        msg += f"\n\nScreen size: {root.winfo_screenwidth()}x{root.winfo_screenheight()}"
    messagebox.showinfo(
        title=f"About: {title()}",
        message=msg,
        parent=root,
    )


def post_right_click_menu(menu: tkinter.Menu, xpos: int, ypos: int):
    """Post menu on arrow position"""

    def destroy(evt):
        menu.unpost()

    try:
        menu.bind('<FocusOut>', destroy)
        menu.tk_popup(xpos, ypos)
    finally:
        menu.grab_release()


def folder_treeview(parent: tk.Misc, columns: list[tuple[str, str, int, bool, str | None]],
                    width: int | None = None, height: int | None = None) -> ttk.Treeview:
    """
    Creates a ttk.TreeView object inside a frame with columns for folders
    """
    canvas = tk.Canvas(parent)
    # fixed size of grid in canvas
    canvas.grid_rowconfigure(0, weight=1)
    canvas.grid_columnconfigure(0, weight=1)
    if width and height:
        canvas.configure(width=width, height=height)
        canvas.pack()
    else:
        canvas.pack(side=tk.TOP, fill=tk.BOTH, expand=tk.YES)
    canvas.grid_propagate(False)

    frm = ttk.Frame(canvas)
    frm.grid_rowconfigure(0, weight=1)
    frm.grid_columnconfigure(0, weight=1)

    tree = ttk.Treeview(frm, columns=[c[0] for c in columns[1:]])
    for c in columns:
        tree.column(c[0], stretch=False)

    var = ttk.Scrollbar(frm, orient="vertical", command=tree.yview)
    # var.pack(side=tk.RIGHT, fill=tk.Y)
    var.grid(column=1, row=0, sticky='nsew')
    tree.configure(yscrollcommand=var.set)

    var = ttk.Scrollbar(frm, orient="horizontal", command=tree.xview)
    # var.pack(side=tk.BOTTOM, fill=tk.X)
    var.grid(column=0, row=1, sticky='ew')
    tree.configure(xscrollcommand=var.set)
    # tree.pack(side=tk.TOP)
    tree.grid(column=0, row=0, sticky='nsew')
    frm.grid(column=0, row=0)

    def tree_sort(col, reverse, sort_col=None):
        return lambda: treeview_sort_column(tree, col, reverse=reverse, sort_col=sort_col)

    for name, text, width, _reverse, _sort_col in columns:
        tree.heading(name, text=text, command=tree_sort(_sort_col or name, _reverse, name if _sort_col else None))
        tree.column(name, width=width, stretch=False)  # stretch stops columns from stretching when resized
    return tree


def capture(root: tk.Misc, filename='img.png'):
    """Take screenshot of the passed widget. Untested!"""

    x0 = root.winfo_rootx()
    y0 = root.winfo_rooty()
    x1 = x0 + root.winfo_width()
    y1 = y0 + root.winfo_height()

    im = ImageGrab.grab(bbox=(x0, y0, x1, y1))  # bbox means boundingbox, which is shown in the image below
    im.save(filename)  # Can also say im.show() to display it


def copy_image_to_clipboard(root: tk.Misc, canvas: tk.Canvas):
    """copy tk canvas to clipboard. Untested!"""

    # https://www.tutorialspoint.com/how-to-copy-a-picture-from-tkinter-canvas-to-clipboard
    import io
    from PIL import Image

    # Retrieve the image from the canvas
    canvas_image = canvas.postscript()

    # Create an in-memory file-like object
    image_buffer = io.BytesIO()

    # Save the canvas image to the buffer in PNG format
    image = Image.open(canvas_image)
    image.save(image_buffer, format="PNG")
    image_buffer.seek(0)

    # Copy the image to the clipboard
    root.clipboard_clear()
    root.clipboard_append(image_buffer, format="image/png")


"""
Various utilities for reading and writing nexus files
"""
import typing

import numpy as np
import h5py

from mmg_toolbox.nexus import nexus_names as nn
from mmg_toolbox.utils.units import METERS, unit_converter
from mmg_toolbox.utils.xray_utils import photon_energy, photon_wavelength


def bytes2str(value: str | bytes | list | tuple) -> str:
    """Convert bytes or string to string"""
    if isinstance(value, (str, bytes)):
        return value.decode('utf-8', errors='ignore') if hasattr(value, 'decode') else value
    return bytes2str(next(iter(value), ''))


def get_attr_datasets(group: h5py.Group, attr_name: str) -> list[h5py.Dataset]:
    """Return list of datasets mentioned in group attribute"""
    attribute = group.attrs.get(attr_name, '')
    if isinstance(attribute, (str, bytes)):
        attribute = [attribute]
    # attribute is a list e.g. @axes=['x', 'y']
    return [dataset for name in attribute if (dataset := group.get(name))]


def reorder_group_items(group: h5py.Group) -> dict[str, h5py.Group | h5py.Dataset]:
    """re-order the group.items list to get the put @default objects first"""
    # 1. put default objects first in the list
    items = {}
    if nn.NX_DEFAULT in group.attrs and group.attrs[nn.NX_DEFAULT] in group:
        items[group.attrs[nn.NX_DEFAULT]] = group[group.attrs[nn.NX_DEFAULT]]
    # 2. put datasets before groups to avoid catching lower-level matches
    items.update({key: ds for key, ds in group.items() if isinstance(ds, h5py.Dataset)})
    # 3. add remaining items
    items.update(group.items())
    return items


def update_args(name: str, obj: h5py.Group, axes: str, signal: str, *args: str) -> list[str]:
    """remove object from search arguments if it matches"""
    # expand match-names with SEARCH_ATTRS
    names = [name] + [bytes2str(obj.attrs.get(attr, '')) for attr in nn.SEARCH_ATTRS]
    # Add NX application definition
    if isinstance(obj, h5py.Group) and nn.NX_DEFINITION in obj:
        names.append(bytes2str(obj[nn.NX_DEFINITION][()]))
    # Add axes & signal from parent group
    if name == axes:
        names.append(nn.NX_AXES)
    if name == signal:
        names.append(nn.NX_SIGNAL)
    # check if object matches first arg, otherwise drill down or continue
    return list(args[1:] if args[0] in names else args)


def nx_find(parent: h5py.Group, *field_or_class: str) -> h5py.Dataset | h5py.Group | None:
    """
    Return default or first object to match a set of NXclass or field names

    Example:
        with h5py.File('/path/to/file.h5', 'r') as hdf
            group = nx_find(hdf, 'NXentry', 'NXinstrument', 'NXdetector')  # returns NXdetector group
            dataset = nx_find(hdf, 'NXdata', 'signal')  # returns @signal dataset in @default NXdata
            data = dataset[()]

    Accepted field_or_class arguments:
        - dataset name, e.g. 'data' (returns dataset)
        - group name, e.g. 'entry' (returns group)
        - NX class name, e.g. 'NX_class' (returns NXclass group)
        - NXclass definition, e.g. 'NXmx' (returns NXentry group)
        - local_name attribute name, e.g. 'eta.eta' (returns dataset)
        - 'axes' or 'signal' in NXdata group (returns dataset)
        - hdf path, e.g. 'group/dataset' (returns dataset)

    Parameters:
    :param parent: parent group, must be h5py.File or h5py.Group
    :param field_or_class: names to search for, in hierarchical order.
    :returns: matching Dataset, Group or None if no match
    """

    #TODO: add tuple option for OR
    def recursor(group: h5py.Group, *args):
        # return object from path, e.g. obj['group/data']
        if len(args) == 1 and args[0] in group:
            return group[args[0]]
        # Get group axes & signal datasets
        axes = bytes2str(group.attrs.get(nn.NX_AXES, ''))
        signal = bytes2str(group.attrs.get(nn.NX_SIGNAL, ''))

        items = reorder_group_items(group)  # @default first
        for name, obj in items.items():
            new_args = update_args(name, obj, axes, signal, *args)
            if len(new_args) == 0:
                return obj
            if isinstance(obj, h5py.Group):
                found_obj = recursor(obj, *new_args)
                if found_obj:
                    return found_obj
        return None
    return recursor(parent, *field_or_class)


def nx_find_all(parent: h5py.Group, *field_or_class: str) -> list[h5py.Dataset | h5py.Group]:
    """
    Return all objects that match a set of NXclass or field names

    Example:
        with h5py.File('/path/to/file.h5', 'r') as hdf
            groups = nx_find_all(hdf, 'NXdetector')  # returns list of NXdetector group
            datasets = nx_find(hdf, 'NXdata', 'signal')  # returns list of @signal datasets in NXdata groups
            arrays = [dataset[()] for dataset in datasets]

    Accepted field_or_class arguments:
        - dataset name, e.g. 'data' (returns dataset)
        - group name, e.g. 'entry' (returns group)
        - NX class name, e.g. 'NX_class' (returns NXclass group)
        - NXclass definition, e.g. 'NXmx' (returns NXentry group)
        - local_name attribute name, e.g. 'eta.eta' (returns dataset)
        - 'axes' or 'signal' in NXdata group (returns dataset)

    Parameters:
    :param parent: parent group, must be h5py.File or h5py.Group
    :param field_or_class: names to search for, in hierarchical order.
    :returns: list of matching Datasets or Groups
    """

    def recursor(group: h5py.Group, *args):
        found = []
        # object from path, e.g. obj['group/data']
        if len(args) == 1 and args[0] in group:
            found.append(group[args[0]])
            return found
        # Get group axes & signal datasets
        axes = bytes2str(group.attrs.get(nn.NX_AXES, ''))
        signal = bytes2str(group.attrs.get(nn.NX_SIGNAL, ''))

        for name, obj in group.items():
            new_args = update_args(name, obj, axes, signal, *args)
            if len(new_args) == 0:
                found.append(obj)
                continue
            if isinstance(obj, h5py.Group):
                found += recursor(obj, *new_args)
        return found
    return recursor(parent, *field_or_class)


def nx_find_data(parent: h5py.Group, *field_or_class: str, default=None):
    """Use nx_find to get dataset, return data or default"""
    dataset = nx_find(parent, *field_or_class)
    if isinstance(dataset, h5py.Dataset):
        if np.issubdtype(dataset.dtype, np.number):
            return dataset[...]
        return dataset.asstr()[...]
    return default


def get_axes_signals(nxdata: h5py.Group) -> tuple[list[h5py.Dataset], list[h5py.Dataset]]:
    """Return lists of axes and signal+auxiliary_signals datasets"""
    axes_datasets = get_attr_datasets(nxdata, nn.NX_AXES)
    signal_datasets = get_attr_datasets(nxdata, nn.NX_SIGNAL)
    aux_datasets = get_attr_datasets(nxdata, nn.NX_AUXILIARY)
    return axes_datasets, signal_datasets + aux_datasets


def get_dataset_string(dataset: h5py.Dataset) -> str:
    """Return formatted string of value stored in dataset"""
    if np.issubdtype(dataset, np.number):
        if dataset.size > 1:
            # numeric ndarray
            return f"{dataset.dtype} {dataset.shape}"
        value = np.squeeze(dataset[()])
        if 'decimals' in dataset.attrs:
            value = value.round(dataset.attrs['decimals'])
        if 'units' in dataset.attrs:
            return f"{value} {bytes2str(dataset.attrs['units'])}"
        return str(value)
    try:
        string_dataset = dataset.asstr()[()]
        if dataset.ndim == 0:
            return str(string_dataset)  # bytes or str -> str
        return f"['{string_dataset[0]}', ...({len(string_dataset)})]"  # str array
    except ValueError:
        return str(np.squeeze(dataset[()]))  # other np.ndarray


def get_metadata(group: h5py.Group, *name_paths_default: tuple[str, tuple, str]) -> dict[str, str]:
    """Return a dict with metadata available in hdf Group. All metadata formated as strings"""
    metadata = {
        name: get_dataset_string(dataset)
        if (dataset := nx_find(group, *paths)) else default
        for name, paths, default in name_paths_default
    }
    return metadata


def get_dataset_value(path: str, group: h5py.Group,
                      default: str | float | np.ndarray | None = None, units: str = '') -> np.ndarray | None:
    """
    Get value from dataset in group, or return default
    :param path: hdf path of dataset in group
    :param group: hdf group
    :param default: returned if path doesn't exist
    :param units: converts to given units if units in attrs
    :return: value or default as ndarray, or None if default is None
    """
    if path in group:
        dataset = group[path]
        if np.issubdtype(dataset, np.number):
            data = np.squeeze(dataset[...])
            if units and 'units' in dataset.attrs:
                return unit_converter(data, bytes2str(dataset.attrs['units']), units)
            return data
        return dataset.asstr()[...]
    return None if default is None else np.asarray(default)


def nx_beam_energy(beam: h5py.Group) -> tuple[float, float]:
    """
    Return beam energy in keV and wavelength in A
    :param beam: Nexus NXbeam group
    :return: incident_energy, incident_wavelength
    """
    if nn.NX_WL in beam:
        dataset = beam[nn.NX_WL]
        units = dataset.attrs.get(nn.NX_UNITS, b'nm').decode()
        wl = dataset[()]
        if units.lower() in METERS:
            wl = wl * METERS[units] * 1e-10  # wavelength in Angstroms
        else:
            print(f"Warning: unknown translation units: {units}")
        return photon_energy(wl), wl
    elif nn.NX_EN in beam:
        dataset = beam[nn.NX_WL]
        units = dataset.attrs.get(nn.NX_UNITS, b'ev').decode()
        en = dataset[()]
        if units.lower() == 'ev':
            en = en / 1000.  # wavelength in keV
        return en, photon_wavelength(en)
    else:
        raise KeyError(f"{beam} contains no '{nn.NX_WL}' or '{nn.NX_EN}'")



"""
Wrapper functions for DiffCalc

Requires:
    pip install diffcalc-core

Example:
    from mmg_toolbox.diffcalc import UB
    ub = UB()
    ub.latt(2.85, 2.85, 10.8, 90, 90, 120)
    ub.add_reflection('ref1', ...)
    ub.add_reflection('ref2', ...)
    ub.calc_ub('ref1', 'ref2')
    ub.con('gamma',0, 'mu',0, 'bisect')

    angles = ub.hkl2angles((1,1,1), energy_kev=6)
    hkl = ub.angles2hkl(phi, chi, eta, mu, delta, gamma, energy_kev)
    solutions = ub.all_solutions((1, 1, 1), energy_kev=6)
"""

import numpy as np
from diffcalc.hkl.calc import HklCalculation
from diffcalc.hkl.constraints import Constraints
from diffcalc.hkl.geometry import Position
from diffcalc.ub.calc import UBCalculation

from mmg_toolbox.utils.xray_utils import photon_wavelength, photon_energy

# Euler to Kappa angles
KALPHA = 50.0 # limits the range of possible chi in different modes.
#Maybe the angle between kth and kappa arms when phi || z ? Or angle needed for chi=90?
CHI_MAGIC = 65.595503 # some unexplained angle


DEFAULT_LATTICE = ('Hexagonal', 2.85, 10.8)
DEFAULT_CONSTRAINTS = {
    'gamma': 0.0,
    'mu': 0.0,
    'phi': 0.0
}
DEFAULT_LIMITS = {
    # angle: (min, max)
    'phi': (-180, 180),
    'chi': (-10, 100),
    'eta': (-20, 120),
    'mu': (-10, 180),
    'delta': (-10, 120),
    'gamma': (-10, 120),
}
RENAME_AXES = {
    # my name: DiffCalc name
    'gamma': 'nu'
}


def _euler2kappa(phi: float, chi: float, eta: float,
                 mode: int = 1, kalpha: float = KALPHA, chi_magic: float = CHI_MAGIC) -> tuple[float, float, float]:
    """
    Convert from Eulerian space angles to real world motor angles
    in: e_angles = [eta, chi, phi] # in degrees
    out : k_angles = [ktheta, kappa, kphi] # in degrees

    This function takes some weird constant parameters, and a mode parameter.
    The code requires mu constraints (mu = 0 or mu = 180 degrees) to work in the specified modes.
    The modes allow to invert the movement of kappa (kappa -> (-1)*kappa) and adjust
    the other angles accordingly.
    The inverted modes could be useful when working with large equipment.

    # Coversion Modes:
    # Mode; Constraint; Effect
    # "  1; mu=0      ; Normal operation
    # "  2; mu=0      ; kappa -> (-1)*kappa
    # "  3; mu=180    ; Normal operation, with mu on the opposite side of the diffractometer.
    phi rotates >180 degrees, be careful with pipes!
    # "  4; mu=180    ; kappa -> (-1)*kappa, with mu on the opposite side.
    phi rotates >180 degrees, be careful with pipes!

    Potential upgrade: Include mu in the calculations so the code could work for any value of mu.

    Known Bugs:
        1. Rotation matrix algebra cannot be varified

    Fixed Bugs (since version 1.0):
        1. Weird behaviour at chi=100.
        2. Errors of accessing undefined k_angle variables
        3. Total ugliness
        4. Added setRange before checking abs(chi)
    """
    cos = lambda x: np.cos(np.deg2rad(x))
    sin = lambda x: np.sin(np.deg2rad(x))
    tan = lambda x: np.tan(np.deg2rad(x))
    asin = lambda x: np.rad2deg(np.arcsin(x))

    def _set_range(value, min_angle: float = -180, max_angle: float = 180):
        if value < min_angle:
            return _set_range(value + 360, min_angle, max_angle)
        elif value > max_angle:
            return _set_range(value - 360, min_angle, max_angle)
        else:
            return value

    # calculates modes 1 and 2 for -100 < chi < 100
    if abs(chi) < 2 * kalpha:
        delta1 = -asin(tan(chi / 2.) / tan(kalpha))
        kappa = -asin(cos(delta1) * sin(chi) / sin(kalpha))

        if abs(chi) > chi_magic and chi > 0.:
            kappa = _set_range(180 - kappa)
        elif abs(chi) > chi_magic and chi < 0.:
            kappa = _set_range(-180 - kappa)

        theta_K1 = _set_range(eta - delta1, -90., 270.)
        phi_K1 = _set_range(phi - delta1, -90., 270.)

        theta_K2 = _set_range(eta - (180 - delta1), -90., 270.)
        phi_K2 = _set_range(phi - (180 - delta1), -90., 270.)
        K2 = _set_range(-kappa)

    # if chi is out of range
    else:
        kappa = None
        K2 = None
        phi_K1 = None
        phi_K2 = None
        theta_K1 = None
        theta_K2 = None

    # calculates modes 3 and 4 for -180 < chi < -100 and 100 < chi < 180
    if abs(chi) > (180. - kalpha * 2):
        chi_r = _set_range(180 - chi)
        delta3 = -asin(tan(chi_r / 2.) / tan(kalpha))
        K3 = -asin(cos(delta3) * sin(chi_r) / sin(kalpha))

        if abs(chi_r) > chi_magic and chi_r > 0.:
            K3 = _set_range(180 - K3)
        elif abs(chi_r) > chi_magic and chi_r < 0.:
            K3 = _set_range(-180 - K3)

        theta_K3 = _set_range(eta - delta3, -90., 270.)
        phi_K3 = _set_range(phi - delta3 + 180, -90., 270.)
        theta_K4 = _set_range(eta - (180 - delta3), -90., 270.)
        phi_K4 = _set_range(phi - (180 - delta3) * 180., -90., 270.)
        K4 = _set_range(-K3)

    # if chi is out of range
    elif abs(chi) <= (180. - kalpha * 2):
        K3 = None
        K4 = None
        phi_K3 = None
        phi_K4 = None
        theta_K3 = None
        theta_K4 = None

    else:
        raise Exception('chi is wrong')

    Kvalues = [
        [theta_K1, kappa, phi_K1],
        [theta_K2, K2, phi_K2],
        [theta_K3, K3, phi_K3],
        [theta_K4, K4, phi_K4]
    ]
    ktheta, kappa, kphi = Kvalues[mode - 1]
    return kphi, kappa, ktheta


def _kappa2euler(kphi: float, kappa: float, ktheta: float,
                 mode: int = 1, kalpha: float = KALPHA) -> tuple[float, float, float]:
    """
    Convert k_angles of real motors to e_angles in Eulerian space
    in : k_angles = [ktheta, kappa, kphi] # in degrees
    out: e_angles = [eta_now, chi_now, phi_now] # in degrees
    mode: must be the same mode as in EtoK()
    """
    cos = lambda x: np.cos(np.deg2rad(x))
    sin = lambda x: np.sin(np.deg2rad(x))
    tan = lambda x: np.tan(np.deg2rad(x))
    asin = lambda x: np.rad2deg(np.arcsin(x))
    atan = lambda x: np.rad2deg(np.arctan(x))

    def _set_range(value, min_angle: float = -180, max_angle: float = 180):
        if value < min_angle:
            return _set_range(value + 360, min_angle, max_angle)
        elif value > max_angle:
            return _set_range(value - 360, min_angle, max_angle)
        else:
            return value

    ktheta = _set_range(ktheta, -90, 270)
    kappa = _set_range(kappa)
    kphi = _set_range(kphi, -90, 270)

    if mode == 1:
        gamma = -atan(cos(kalpha) * tan(kappa / 2.))
        chi = -2 * asin(sin(kappa / 2) * sin(kalpha))
        theta = ktheta - gamma
        phi = kphi - gamma

    elif mode == 2:
        gamma = -atan(cos(kalpha) * tan(kappa / 2)) + 180.
        chi = 2 * asin(sin(kappa / 2) * sin(kalpha))
        theta = ktheta - gamma
        phi = kphi - gamma

    elif mode == 3:
        gamma = -atan(cos(kalpha) * tan(kappa / 2.))
        chi = 2 * asin(sin(kappa / 2) * sin(kalpha)) + 180.
        theta = ktheta - gamma
        phi = kphi - gamma + 180.

    elif mode == 4:
        gamma = -atan(cos(kalpha) * tan(kappa / 2.))
        chi = -2 * asin(sin(kappa / 2) * sin(kalpha)) + 180.
        theta = ktheta - gamma + 180.
        phi = kphi - gamma
        theta = _set_range(theta, -90., 270.)
        chi = _set_range(chi)
        phi = _set_range(phi, -90., 270.)

    else:
        raise Exception('mode not recognized')

    eta, chi, phi = _set_range(theta, -90., 270.), _set_range(chi), _set_range(phi, -90., 270.)
    return phi, chi, eta


def euler2kappa(phi: np.ndarray, chi: np.ndarray, eta: np.ndarray,
                mode: int = 1, kalpha: float = KALPHA, chi_magic: float = CHI_MAGIC
                ) -> tuple[np.ndarray, np.ndarray, np.ndarray]:
    phi = np.reshape(phi, -1)
    chi = np.reshape(chi, -1)
    eta = np.reshape(eta, -1)
    kphi, kappa, ktheta = np.array([
        _euler2kappa(p, c, e,  mode=mode, kalpha=kalpha, chi_magic=chi_magic)
        for p, c, e in zip(phi, chi, eta)
    ]).T
    return kphi.squeeze(), kappa.squeeze(), ktheta.squeeze()


def kappa2euler(kphi: np.ndarray, kappa: np.ndarray, ktheta: np.ndarray,
                mode: int = 1, kalpha: float = KALPHA,
                ) -> tuple[np.ndarray, np.ndarray, np.ndarray]:
    kphi = np.reshape(kphi, -1)
    kappa = np.reshape(kappa, -1)
    ktheta = np.reshape(ktheta, -1)
    phi, chi, eta = np.array([
        _kappa2euler(kph, kap, kth,  mode=mode, kalpha=kalpha)
        for kph, kap, kth in zip(kphi, kappa, ktheta)
    ]).T
    return phi.squeeze(), chi.squeeze(), eta.squeeze()


class UB:
    """
    Wrapper class for DiffCalc functionality
    """
    ubcalc: UBCalculation
    constrains: Constraints
    limits: dict[str, tuple[float, float]]
    hklcalc: HklCalculation

    def __init__(self,
                 lattice: tuple[float, float, float, float, float, float] | None = None,
                 ref1: tuple[tuple[int, int, int], float, float, float, float, float, float, float] | None = None,
                 ref2: tuple[tuple[int, int, int], float, float, float, float, float, float, float] | None = None,
                 constraints: dict[str, float | None] | None = None,
                 limits: dict[str, tuple[float, float]] | None = None,
                 azir: tuple[float, float, float] = (1, 0, 0),
                 surface: tuple[float, float, float] = (0, 0, 1),
                 ):
        self.ubcalc = UBCalculation("sixcircle")
        if lattice is None:
            lattice = DEFAULT_LATTICE
        self.ubcalc.set_lattice("sample", *lattice)
        self.ubcalc.n_hkl = azir # azimuthal reference, hkl
        self.ubcalc.surf_nhkl = surface  # surface vector, hkl
        if ref1 is not None:
            self.add_reflection('ref1', *ref1)
        if ref2 is not None:
            self.add_reflection('ref2', *ref2)
        if ref1 is not None and ref2 is not None:
            self.ubcalc.calc_ub('ref1', 'ref2')

        if constraints is None:
            constraints = DEFAULT_CONSTRAINTS.copy()
        self.constraints = Constraints(self._names2diffcalc(**constraints))
        if limits is None:
            limits = DEFAULT_LIMITS.copy()
        self.limits = limits
        self.hklcalc = HklCalculation(self.ubcalc, self.constraints)
        self.lab_transformation = np.eye(3)

    def __repr__(self):
        return repr(self.ubcalc)

    def __str__(self):
        return f"{self.ubcalc}\n\nCONSTRAINTS\n{self.constraints}"

    def add_reflection(self, label: str, hkl: tuple[float, float, float],
                       phi: float = 0, chi: float = 0, eta: float = 0, mu: float = 0,
                       delta: float = 0, gamma: float = 0,
                       energy_kev: float | None = None, wavelength_a: float | None = None):
        """
        Add a reflection location to the UB calculation
        """
        if energy_kev is None:
            energy_kev = photon_energy(wavelength_a)
        self.ubcalc.add_reflection(
            tag=label,
            hkl=hkl,
            position=Position(
                indegrees=True,
                phi=phi,
                chi=chi,
                eta=eta,
                mu=mu,
                delta=delta,
                nu=gamma,
            ),
            energy=energy_kev,
        )

    def add_orientation(self, label: str, hkl: tuple[float, float, float], xyz: tuple[float, float, float]):
        """
        Add an orientation to the UB calculation
        """
        self.ubcalc.add_orientation(
            # Add orientation of crystal, can be used together with reflection
            hkl=hkl,
            xyz=xyz,  # xyz in diffractometer basis
            position=None,
            tag=label
        )

    def latt(self, a: float = 5.0, b: float | None = None, c: float | None = None,
             alpha: float | None = None, beta: float | None = None, gamma: float | None = None,
             name: str = 'xtl', system: str = None):
        """
        Set the crystal lattice
        """
        self.ubcalc.set_lattice(name, system, a, b, c, alpha, beta, gamma)

    def calcub(self, tag1: str = 'ref1', tag2: str = 'ref2'):
        self.ubcalc.calc_ub(tag1, tag2)

    def set_lab_transformation(self, transformation: np.ndarray):
        """Assign a 3x3 transformation matrix to appy to angles"""
        self.lab_transformation = transformation

    def ub_matrix(self):
        return np.dot(self.lab_transformation, self.ubcalc.UB)

    def orientation_matrix(self):
        return np.dot(self.lab_transformation, self.ubcalc.U)

    def lp(self):
        xtl = self.ubcalc.crystal
        return xtl.a1, xtl.a2, xtl.a3, xtl.alpha1, xtl.alpha2, xtl.alpha3

    def asdict(self):
        return self.hklcalc.asdict.copy()

    def load_from_diffcalc(self, hklcalc: HklCalculation | None = None,
                           ubcalc: UBCalculation | None = None,
                           constraints: Constraints | None = None):
        if hklcalc is not None:
            self.hklcalc = hklcalc
            self.ubcalc = hklcalc.ubcalc
            self.constraints = hklcalc.constraints
        else:
            if ubcalc is not None:
                self.ubcalc = ubcalc
            if constraints is not None:
                self.constraints = constraints

    def azir(self, hkl: tuple[float, float, float]):
        """Set azimuthal reference reflection"""
        self.ubcalc.n_hkl = hkl

    def _names2diffcalc(self, **kwargs):
        return {
            RENAME_AXES[name] if name in RENAME_AXES else name: value
            for name, value in kwargs.items()
        }

    def _diffcalc2names(self, **kwargs):
        RENAME = {value: name for name, value in RENAME_AXES.items()}
        return {
            RENAME[name] if name in RENAME else name: value
            for name, value in kwargs.items()
        }

    def set_limits(self, **limits: [str, tuple[float, float]]):
        """Set limits"""
        limits = self._names2diffcalc(**limits)
        self.limits.update(limits)

    def _set_constraints(self, **constraints):
        """Set constraints"""
        constraints = self._names2diffcalc(**constraints)
        self.constraints.clear()
        for name, value in constraints.items():
            setattr(self.constraints, name, value)

    def con(self,
            con1: str, value1: float | bool,
            con2: str, value2: float | bool,
            con3: str, value3: float | bool = True):
        """Apply contraints to the DiffCalc calculation"""
        self._set_constraints(**{con1: value1, con2: value2, con3: value3})

    def bisect_vertical(self):
        self.con('gamma', 0, 'mu', 0, 'bisect')

    def bisect_horizontal(self):
        self.con('delta', 0, 'eta', 0, 'bisect')

    def fixed_phi_vertical(self, phi: float = 0):
        self.con('gamma', 0, 'mu', 0, 'phi', phi)

    def fixed_phi_horizontal(self, phi: float = 0):
        self.con('delta', 0, 'eta', 0, 'phi', phi)

    def fixed_psi_vertical(self, psi: float = 0):
        self.con('gamma', 0, 'mu', 0, 'psi', psi)

    def fixed_psi_horizontal(self, psi: float = 0):
        self.con('delta', 0, 'eta', 0, 'psi', psi)

    def all_solutions(self, hkl: tuple[float, float, float],
                      energy_kev: float | None = None, wavelength_a: float | None = None) -> list[dict[str, float]]:
        """Calculate all angle solutions for the given HKL"""
        if wavelength_a is None:
            wavelength_a = photon_wavelength(energy_kev)
        h, k, l = hkl
        all_pos = self.hklcalc.get_position(h, k, l, wavelength_a)
        solutions = []
        for posn, virtual_angles in all_pos:
            pos = posn.asdict.copy()
            pos['gamma'] = pos['nu']
            ktheta, kappa, kphi = euler2kappa(pos['phi'], pos['chi'], pos['eta'])
            kap = {
                'ktheta': ktheta,
                'kappa': kappa,
                'kphi': kphi
            }
            solutions.append(self._diffcalc2names(**{**pos, **virtual_angles, **kap}))
        return solutions

    def hkl2angles(self, hkl: tuple[float, float, float],
                   energy_kev: float | None = None, wavelength_a: float | None = None) -> dict[str, float] | None:
        """Calculate the angles for the given HKL"""
        if wavelength_a is None:
            wavelength_a = photon_wavelength(energy_kev)
        h, k, l = hkl
        all_pos = self.hklcalc.get_position(h, k, l, wavelength_a)
        for posn, virtual_angles in all_pos:
            pos = posn.asdict.copy()
            pos['gamma'] = pos['nu']
            if all(
                ax_min < pos.get(axis) < ax_max for axis, (ax_min, ax_max) in self.limits.items()
            ):
                ktheta, kappa, kphi = euler2kappa(pos['phi'], pos['chi'], pos['eta'])
                kap = {
                    'ktheta': ktheta,
                    'kappa': kappa,
                    'kphi': kphi
                }
                return self._diffcalc2names(**{**pos, **virtual_angles, **kap})
        return None

    def angles2hkl(self, phi: float = 0, chi: float = 0, eta: float = 0, mu: float = 0,
                   delta: float = 0, gamma: float = 0,
                   energy_kev: float | None = None, wavelength_a: float | None = None) -> tuple[float, float, float]:
        """Calculate the HKL for the given angles"""
        if wavelength_a is None:
            wavelength_a = photon_wavelength(energy_kev)
        pos = Position(
            # Eulerean angles in You. et al diffractometer basis
            # (z-along phi when all angles 0, y along beam, x vertical)
            nu=gamma,  # detector rotation, positive about diffractometer x-axis (gamma)
            delta=delta,  # detector rotation, negative about diffractometer z-axis
            mu=mu,  # sample rotation, positive about diffractometer x-axis
            eta=eta,  # sample rotation, negative about diffractometer z-axis
            chi=chi,  # sample rotation, positive about diffractometer y-axis
            phi=phi  # sample rotation, negative about diffractometer z-axis
        )
        return self.hklcalc.get_hkl(pos, wavelength_a)

    def kappa2hkl(self, ktheta: float = 0, kappa: float = 0, kphi: float = 0, mu: float = 0,
                  delta: float = 0, gamma: float = 0,
                  energy_kev: float | None = None, wavelength_a: float | None = None) -> tuple[float, float, float]:
        """Calculate the HKL for the given Kappa-angles"""
        if wavelength_a is None:
            wavelength_a = photon_wavelength(energy_kev)
        phi, chi, eta = kappa2euler(ktheta, kappa, kphi)
        pos = Position(nu=gamma, delta=delta, mu=mu, eta=eta, chi=chi, phi=phi)
        return self.hklcalc.get_hkl(pos, wavelength_a)

    def euler2kappa(self, phi: float = 0, chi: float = 0, eta: float = 0) -> tuple[float, float, float]:
        """Calculate the Euler angles for the given Kappa-angles"""
        return euler2kappa(phi, chi, eta)

    def kappa2euler(self, ktheta: float = 0, kappa: float = 0, kphi: float = 0) -> tuple[float, float, float]:
        """Calculate the Euler angles for the given Kappa-angles"""
        return kappa2euler(ktheta, kappa, kphi)

